from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Generic, Literal, Self, TypeAlias, TypeVar

from whenever import Instant

from hassette.core import topics
from hassette.core.events import Event
from hassette.models.states import StateT, try_convert_state

from .raw import HassEventEnvelopeDict, HassStateDict

HassT = TypeVar("HassT", covariant=True)
MISSING_VALUE = object()  # Used to indicate a missing value in states


@dataclass(slots=True, frozen=True)
class HassContext:
    """Structure for the context of a state change event."""

    id: str
    parent_id: str | None
    user_id: str | None


@dataclass(slots=True, frozen=True)
class HassPayload(Generic[HassT]):
    """Base class for Home Assistant event payloads."""

    event_type: str
    data: HassT
    origin: Literal["LOCAL", "REMOTE"]
    time_fired: Instant
    context: HassContext

    @property
    def entity_id(self) -> str | None:
        """Return the entity ID if present in the data."""
        return getattr(self.data, "entity_id", None)

    @property
    def domain(self) -> str | None:
        """Return the domain if present in the data."""
        if hasattr(self.data, "domain"):
            return getattr(self.data, "domain", None)

        entity_id = self.entity_id
        if entity_id:
            return entity_id.split(".")[0]
        return None

    @property
    def service(self) -> str | None:
        """Return the service if present in the data."""
        return getattr(self.data, "service", None)


@dataclass(slots=True, frozen=True)
class CallServicePayload:
    domain: str
    service: str
    service_data: dict[str, Any] = field(default_factory=dict)
    service_call_id: str | None = None  # have never seen this but the docs say it exists


@dataclass(slots=True, frozen=True)
class ComponentLoadedPayload:
    component: str


@dataclass(slots=True, frozen=True)
class ServiceRegisteredPayload:
    domain: str
    service: str


@dataclass(slots=True, frozen=True)
class ServiceRemovedPayload:
    domain: str
    service: str


@dataclass(slots=True, frozen=True)
class LogbookEntryPayload:
    name: str
    message: str
    domain: str | None = None
    entity_id: str | None = None


@dataclass(slots=True, frozen=True)
class UserAddedPayload:
    user_id: str


@dataclass(slots=True, frozen=True)
class UserRemovedPayload:
    user_id: str


@dataclass(slots=True, frozen=True)
class AutomationTriggeredPayload:
    name: str
    entity_id: str
    source: str  # this one isn't on the docs page but is included apparently
    # https://www.home-assistant.io/docs/configuration/events/#automation_triggered


@dataclass(slots=True, frozen=True)
class ScriptStartedPayload:
    name: str
    entity_id: str


@dataclass(slots=True, frozen=True)
class StateChangePayload(Generic[StateT]):
    entity_id: str
    old_state: None | StateT
    """The previous state of the entity before it changed. Omitted if the state is set for the first time."""

    new_state: None | StateT
    """The new state of the entity. Omitted if the state has been removed."""

    @property
    def state_value_has_changed(self) -> bool:
        """Check if the state value has changed between old and new states.

        Appropriately handles cases where either state may be None.

        Returns:
            bool: True if the state value has changed, False otherwise.
        """
        return self.old_state_value != self.new_state_value

    @property
    def new_state_value(self) -> Any:
        """Return the value of the new state, or MISSING_VALUE if not present."""
        return self.new_state.value if self.new_state is not None else MISSING_VALUE

    @property
    def old_state_value(self) -> Any:
        """Return the value of the old state, or MISSING_VALUE if not present."""
        return self.old_state.value if self.old_state is not None else MISSING_VALUE

    @property
    def has_new_state(self) -> bool:
        """Check if the new state is not None."""
        return self.new_state is not None

    @property
    def has_old_state(self) -> bool:
        """Check if the old state is not None."""
        return self.old_state is not None

    @classmethod
    def create_from_event(
        cls, entity_id: str, old_state: HassStateDict | None = None, new_state: HassStateDict | None = None
    ) -> Self:
        if entity_id is None:
            raise ValueError("State change event data must contain 'entity_id' key")

        # use deepcopy to avoid mutating the original data
        old_state_obj = try_convert_state(deepcopy(old_state)) if old_state else None
        new_state_obj = try_convert_state(deepcopy(new_state)) if new_state else None

        return cls(entity_id=entity_id, old_state=old_state_obj, new_state=new_state_obj)  # pyright: ignore[reportArgumentType]


def create_event_from_hass(
    data: "HassEventEnvelopeDict",
):
    """Create an Event from a dictionary."""

    from hassette.core.events import Event  # avoid circular import

    event = data.get("event", {})
    event_type = event.get("event_type")
    if not event_type:
        raise ValueError("Event data must contain 'event_type' key")

    event_data = event.get("data", {}) or {}
    event_payload = {
        "event_type": event_type,
        "context": event["context"],
        "origin": event["origin"],
        "time_fired": event["time_fired"],
    }

    if event_type == "state_changed":
        return Event(
            topic=topics.HASS_EVENT_STATE_CHANGED,
            payload=HassPayload(**event_payload, data=StateChangePayload.create_from_event(**event_data)),
        )

    match event_type:
        case "call_service":
            payload_cls = CallServicePayload
            topic = topics.HASS_EVENT_CALL_SERVICE
        case "component_loaded":
            payload_cls = ComponentLoadedPayload
            topic = topics.HASS_EVENT_COMPONENT_LOADED
        case "service_registered":
            payload_cls = ServiceRegisteredPayload
            topic = topics.HASS_EVENT_SERVICE_REGISTERED
        case "service_removed":
            payload_cls = ServiceRemovedPayload
            topic = topics.HASS_EVENT_SERVICE_REMOVED
        case "logbook_entry":
            payload_cls = LogbookEntryPayload
            topic = topics.HASS_EVENT_LOGBOOK_ENTRY
        case "user_added":
            payload_cls = UserAddedPayload
            topic = topics.HASS_EVENT_USER_ADDED
        case "user_removed":
            payload_cls = UserRemovedPayload
            topic = topics.HASS_EVENT_USER_REMOVED
        case "automation_triggered":
            payload_cls = AutomationTriggeredPayload
            topic = topics.HASS_EVENT_AUTOMATION_TRIGGERED
        case "script_started":
            payload_cls = ScriptStartedPayload
            topic = topics.HASS_EVENT_SCRIPT_STARTED
        case _:
            payload_cls = dict
            topic = f"hass.event.{event_type}"

    if payload_cls:
        return Event(topic=topic, payload=HassPayload(**event_payload, data=payload_cls(**event_data)))

    raise ValueError(f"Unknown event type: {event_type}")


StateChangeEvent: TypeAlias = Event[HassPayload[StateChangePayload[StateT]]]
CallServiceEvent: TypeAlias = Event[HassPayload[CallServicePayload]]
ComponentLoadedEvent: TypeAlias = Event[HassPayload[ComponentLoadedPayload]]
ServiceRegisteredEvent: TypeAlias = Event[HassPayload[ServiceRegisteredPayload]]
ServiceRemovedEvent: TypeAlias = Event[HassPayload[ServiceRemovedPayload]]
LogbookEntryEvent: TypeAlias = Event[HassPayload[LogbookEntryPayload]]
UserAddedEvent: TypeAlias = Event[HassPayload[UserAddedPayload]]
UserRemovedEvent: TypeAlias = Event[HassPayload[UserRemovedPayload]]
AutomationTriggeredEvent: TypeAlias = Event[HassPayload[AutomationTriggeredPayload]]
ScriptStartedEvent: TypeAlias = Event[HassPayload[ScriptStartedPayload]]

HassEvent: TypeAlias = Event[HassPayload[Any]]



from typing import Literal

from pydantic import Field
from whenever import Instant, PlainDateTime, ZonedDateTime

from .base import AttributesBase, StringBaseState


class CalendarState(StringBaseState):
    class Attributes(AttributesBase):
        message: str | None = Field(default=None)
        all_day: bool | None = Field(default=None)
        start_time: Instant | PlainDateTime | ZonedDateTime | None = Field(default=None)
        end_time: Instant | PlainDateTime | ZonedDateTime | None = Field(default=None)
        location: str | None = Field(default=None)
        description: str | None = Field(default=None)
        offset_reached: bool | None = Field(default=None)

    domain: Literal["calendar"]

    attributes: Attributes



from typing import Any, Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class UpdateState(StringBaseState):
    class Attributes(AttributesBase):
        auto_update: bool | None = Field(default=None)
        display_precision: int | float | None = Field(default=None)
        installed_version: str | None = Field(default=None)
        in_progress: bool | None = Field(default=None)
        latest_version: str | None = Field(default=None)
        release_summary: Any | None = Field(default=None)
        release_url: str | None = Field(default=None)
        skipped_version: Any | None = Field(default=None)
        title: str | None = Field(default=None)
        update_percentage: Any | None = Field(default=None)
        entity_picture: str | None = Field(default=None)

    domain: Literal["update"]

    attributes: Attributes



import typing

from aiohttp import web

from hassette.core.classes import Service
from hassette.core.enums import ResourceStatus

if typing.TYPE_CHECKING:
    from hassette.core.core import Hassette


class _HealthService(Service):
    """Tiny HTTP server exposing /healthz for container healthchecks."""

    def __init__(self, hassette: "Hassette", host: str = "0.0.0.0", port: int | None = None):
        super().__init__(hassette)
        self.host = host
        self.port = port or hassette.config.health_service_port

        self._runner: web.AppRunner | None = None

    async def run_forever(self) -> None:
        if not self.hassette.config.run_health_service:
            self.logger.info("Health service disabled by configuration")
            return

        try:
            app = web.Application()
            try:
                hassette_key = web.AppKey["Hassette"]("hassette")
            except UnboundLocalError:
                # seen in tests, unsure if it happens in production
                hassette_key = "hassette"

            app[hassette_key] = self.hassette
            app.router.add_get("/healthz", self._handle_health)

            self._runner = web.AppRunner(app)
            await self._runner.setup()
            site = web.TCPSite(self._runner, self.host, self.port)
            await site.start()

            self.logger.info("Health service listening on %s:%s", self.host, self.port)

            # don't send start event until server is running
            await self.handle_start()
            # Just idle until cancelled
            await self.hassette._shutdown_event.wait()
        except OSError as e:
            error_no = e.errno if hasattr(e, "errno") else type(e)
            self.logger.error("Health service failed to start: %s (errno=%s)", e, error_no)
            await self.handle_failed(e)
            raise
        except Exception as e:
            await self.handle_crash(e)
            raise
        finally:
            await self._cleanup()

    async def _cleanup(self) -> None:
        if self._runner:
            await self._runner.cleanup()
            self.logger.debug("Health service stopped")
        if self.status != ResourceStatus.STOPPED:
            await self.handle_stop()

    async def _handle_health(self, request: web.Request) -> web.Response:
        # You can check internals here (e.g., WS status)
        ws_running = self.hassette._websocket.status == ResourceStatus.RUNNING
        if ws_running:
            self.logger.debug("Health check OK")
            return web.json_response({"status": "ok", "ws": "connected"})
        self.logger.warning("Health check FAILED: WebSocket disconnected")
        return web.json_response({"status": "degraded", "ws": "disconnected"}, status=503)



from typing import Literal

from pydantic import Field

from hassette.const.sensor import UNIT_OF_MEASUREMENT

from .base import AttributesBase, NumericBaseState


class NumberState(NumericBaseState):
    class Attributes(AttributesBase):
        min: int | float | None = Field(default=None)
        max: int | float | None = Field(default=None)
        step: int | float | None = Field(default=None)
        mode: str | None = Field(default=None)
        unit_of_measurement: UNIT_OF_MEASUREMENT | str | None = Field(default=None)

    domain: Literal["number"]

    attributes: Attributes



import typing
from collections.abc import Awaitable
from typing import Any, Protocol, TypeVar

from whenever import SystemDateTime

if typing.TYPE_CHECKING:
    from hassette.core.events import Event

E_contra = TypeVar("E_contra", bound="Event[Any]", contravariant=True)


class Predicate(Protocol[E_contra]):
    """Protocol for defining predicates that evaluate events."""

    def __call__(self, event: E_contra) -> bool | Awaitable[bool]: ...


class Handler(Protocol[E_contra]):
    """Protocol for defining event handlers."""

    def __call__(self, event: E_contra) -> Awaitable[None] | None: ...


class AsyncHandler(Protocol[E_contra]):
    """Protocol for defining asynchronous event handlers."""

    def __call__(self, event: E_contra) -> Awaitable[None]: ...


class TriggerProtocol(Protocol):
    def next_run_time(self) -> SystemDateTime:
        """Return the next run time of the trigger."""
        ...



from .classes import Resource
from .events import HassetteServiceEvent


class _ServiceWatcher(Resource):
    """Watches for service events and handles them."""

    async def initialize(self, *args, **kwargs) -> None:
        self._register_internal_event_listeners()
        return await super().initialize(*args, **kwargs)

    async def restart_service(self, event: HassetteServiceEvent) -> None:
        """Start a service from a service event."""
        data = event.payload.data
        name = data.resource_name
        role = data.role

        try:
            if name is None:
                self.logger.warning("No %s specified to start, skipping", role)
                return

            self.logger.info("%s '%s' is being restarted after '%s'", role, name, event.payload.event_type)

            self.logger.info("Starting %s '%s'", role, name)
            service = self.hassette._resources.get(name)
            if service is None:
                self.logger.warning("No %s found for '%s', skipping start", role, name)
                return

            service.cancel()
            service.start()

        except Exception as e:
            self.logger.error("Failed to restart %s '%s': %s", role, name, e)
            raise

    async def log_service_event(self, event: HassetteServiceEvent) -> None:
        """Log the startup of a service."""

        name = event.payload.data.resource_name
        role = event.payload.data.role

        if name is None:
            self.logger.warning("No resource specified for startup, cannot log")
            return

        status, previous_status = event.payload.data.status, event.payload.data.previous_status

        if status == previous_status:
            self.logger.debug("%s '%s' status unchanged at '%s', not logging", role, name, status)
            return

        try:
            self.logger.info(
                "%s '%s' transitioned to status '%s' from '%s'",
                role,
                name,
                event.payload.data.status,
                event.payload.data.previous_status,
            )

        except Exception as e:
            self.logger.error("Failed to log %s startup for '%s': %s", role, name, e)
            raise

    async def shutdown_if_crashed(self, event: HassetteServiceEvent) -> None:
        """Shutdown the Hassette instance if a service has crashed."""
        data = event.payload.data
        name = data.resource_name
        role = data.role

        try:
            self.logger.exception(
                "%s '%s' has crashed (event_id %d), shutting down Hassette, %s",
                role,
                name,
                data.event_id,
                data.exception_traceback,
            )
            self.hassette.shutdown()
        except Exception:
            self.logger.error("Failed to handle %s crash for '%s': %s", role, name)
            raise

    def _register_internal_event_listeners(self) -> None:
        """Register internal event listeners for resource lifecycle."""
        self.hassette.bus.on_hassette_service_failed(handler=self.restart_service)
        self.hassette.bus.on_hassette_service_crashed(handler=self.shutdown_if_crashed)
        self.hassette.bus.on_hassette_service_status(handler=self.log_service_event)



from pathlib import Path

from watchfiles import awatch

from hassette import Service
from hassette.core.events.hassette import create_file_watcher_event


class _FileWatcher(Service):
    """Background task to watch for file changes and reload apps."""

    async def run_forever(self) -> None:
        """Watch app directories for changes and trigger reloads."""
        try:
            self.logger.info("Starting file watcher service")

            paths = self.hassette.config.get_watchable_files()

            self.logger.info("Watching app directories for changes: %s", ", ".join(str(p) for p in paths))

            await self.handle_start()
            async for changes in awatch(
                *paths,
                stop_event=self.hassette._shutdown_event,
                step=self.hassette.config.file_watcher_step_milliseconds,
                debounce=self.hassette.config.file_watcher_debounce_milliseconds,
            ):
                if self.hassette._shutdown_event.is_set():
                    break

                for _, changed_path in changes:
                    changed_path = Path(changed_path).resolve()
                    self.logger.info("Detected change in %s", changed_path)
                    event = create_file_watcher_event(changed_file_path=changed_path)
                    await self.hassette.send_event(event.topic, event)

                # update paths in case new apps were added
                paths = self.hassette.config.get_watchable_files()

        except Exception as e:
            self.logger.exception("App watcher encountered an error, exception args: %s", e.args)
            await self.handle_crash(e)
            raise



import logging

from .config import HassetteConfig
from .core import (
    Api,
    App,
    AppConfig,
    AppConfigT,
    AppSync,
    AsyncHandler,
    CronTrigger,
    Handler,
    HomeAssistantRestarted,
    IntervalTrigger,
    Not,
    Resource,
    ResourceRole,
    ResourceStatus,
    ScheduledJob,
    Service,
    TriggerProtocol,
    events,
    topics,
)
from .core.bus import predicates
from .core.events import StateChangeEvent
from .models import entities, states

logging.getLogger("hassette").addHandler(logging.NullHandler())

__all__ = [
    "Api",
    "App",
    "AppConfig",
    "AppConfigT",
    "AppSync",
    "AsyncHandler",
    "CronTrigger",
    "Handler",
    "HassetteConfig",
    "HomeAssistantRestarted",
    "IntervalTrigger",
    "Not",
    "Resource",
    "ResourceRole",
    "ResourceStatus",
    "ScheduledJob",
    "Service",
    "StateChangeEvent",
    "TriggerProtocol",
    "entities",
    "events",
    "predicates",
    "states",
    "topics",
]



from typing import Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class ClimateState(StringBaseState):
    class Attributes(AttributesBase):
        hvac_modes: list[str] | None = Field(default=None)
        min_temp: int | float | None = Field(default=None)
        max_temp: int | float | None = Field(default=None)
        fan_modes: list[str] | None = Field(default=None)
        preset_modes: list[str] | None = Field(default=None)
        current_temperature: int | float | None = Field(default=None)
        temperature: int | float | None = Field(default=None)
        target_temp_high: float | None = Field(default=None)
        target_temp_low: float | None = Field(default=None)
        current_humidity: float | None = Field(default=None)
        fan_mode: str | None = Field(default=None)
        hvac_action: str | None = Field(default=None)
        preset_mode: str | None = Field(default=None)
        swing_mode: str | None = Field(default=None)
        swing_modes: list[str] | None = Field(default=None)

    domain: Literal["climate"]

    attributes: Attributes



from typing import Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class FanState(StringBaseState):
    class Attributes(AttributesBase):
        preset_modes: list[str] | None = Field(default=None)
        oscillating: bool | None = Field(default=None)
        percentage: int | float | None = Field(default=None)
        percentage_step: float | None = Field(default=None)
        preset_mode: str | None = Field(default=None)
        temperature: int | float | None = Field(default=None)
        model: str | None = Field(default=None)
        sn: str | None = Field(default=None)
        screen_status: bool | None = Field(default=None)
        child_lock: bool | None = Field(default=None)
        night_light: str | None = Field(default=None)
        mode: str | None = Field(default=None)

    domain: Literal["fan"]

    attributes: Attributes



import asyncio
import json
import logging
import typing
from contextlib import AsyncExitStack, suppress
from dataclasses import dataclass
from itertools import count
from logging import getLogger
from typing import Any, cast

import aiohttp
import anyio
import tenacity
from aiohttp import ClientConnectorError, ClientOSError, ClientTimeout, ServerDisconnectedError, WSMsgType
from aiohttp.client_exceptions import ClientConnectionResetError
from tenacity import (
    AsyncRetrying,
    before_sleep_log,
    retry_if_exception_type,
    retry_if_not_exception_type,
)

from hassette.core.classes import Service
from hassette.core.enums import ResourceStatus
from hassette.core.events import HassEventEnvelopeDict, create_event_from_hass, create_websocket_status_event
from hassette.exceptions import (
    ConnectionClosedError,
    CouldNotFindHomeAssistantError,
    FailedMessageError,
    InvalidAuthError,
    RetryableConnectionClosedError,
)

if typing.TYPE_CHECKING:
    from hassette.core.core import Hassette

LOGGER = getLogger(__name__)


@dataclass(frozen=True, eq=True)
class FailedMessagePayload:
    id: int
    error: str


class _Websocket(Service):
    def __init__(self, hassette: "Hassette"):
        super().__init__(hassette)
        self.hassette = hassette
        self.url = self.hassette.config.ws_url
        self._stack = AsyncExitStack()
        self._session: aiohttp.ClientSession | None = None
        self._ws: aiohttp.ClientWebSocketResponse | None = None
        self._response_futures: dict[int, asyncio.Future[Any]] = {}
        self._seq = count(1)

        self._recv_task: asyncio.Task | None = None
        self._subscription_ids: set[int] = set()
        self._connect_lock = asyncio.Lock()  # if you don't already have it

    @property
    def timeout_seconds(self) -> int:
        return self.hassette.config.websocket_timeout_seconds

    @property
    def connected(self) -> bool:
        if self._ws is None:
            return False

        if self._ws._conn is None:
            return False

        return not self._ws._conn.closed

    def get_next_message_id(self) -> int:
        """Get the next message ID."""
        return next(self._seq)

    async def run_forever(self) -> None:
        async with self._connect_lock:
            try:
                await self._connect_and_run()
            except (RetryableConnectionClosedError, ServerDisconnectedError, ClientConnectorError, ClientOSError) as e:
                await self._send_connection_lost_event(f"{type(e).__name__}: {e}")
                await self.handle_failed(e)
            except asyncio.CancelledError:
                await self._send_connection_lost_event("Connection cancelled")
                await self.handle_stop()
            except Exception as e:
                await self.handle_crash(e)
                raise
            finally:
                await self._cleanup()

    async def _connect_and_run(self) -> None:
        """Connect to the WebSocket and run the receive loop."""

        # we wait_fixed for 1 second and try 60 times
        # this lets us (hopefully) get subscribed ASAP after HA is back up
        # instead of waiting exponential and potentially being seconds behind

        # the reason we want to be ASAP is so we can use events like homeassistant_start(ed) to
        # adjust behavior in things like the Api class

        timeout = ClientTimeout(connect=5, total=30)
        async for attempt in AsyncRetrying(
            retry=retry_if_not_exception_type(
                (InvalidAuthError, asyncio.CancelledError, CouldNotFindHomeAssistantError)
            )
            | retry_if_exception_type(
                (RetryableConnectionClosedError, ServerDisconnectedError, ClientConnectorError, ClientOSError)
            ),
            wait=tenacity.wait_fixed(1),
            stop=tenacity.stop_after_attempt(60),
            reraise=True,
            before_sleep=before_sleep_log(LOGGER, logging.WARNING),
        ):
            with attempt:
                async with aiohttp.ClientSession(timeout=timeout) as session:
                    self._session = session
                    try:
                        self._ws = await session.ws_connect(self.url, heartbeat=30)
                    except ClientConnectorError as exc:
                        if exc.__cause__ and isinstance(exc.__cause__, ConnectionRefusedError):
                            raise CouldNotFindHomeAssistantError(self.url) from exc.__cause__
                        raise

                    self.logger.debug("Connected to WebSocket at %s", self.url)
                    await self.authenticate()

                    # start the reader first so send_and_wait can get replies
                    self._recv_task = self.hassette.create_task(self._recv_loop())

                    sub_all_id = await self._subscribe_events()  # uses send_and_wait internally
                    self._subscription_ids.add(sub_all_id)

                    await self.handle_start()
                    event = create_websocket_status_event(connected=True, url=self.url)
                    await self.hassette.send_event(event.topic, event)

                    # Keep running until recv loop ends (disconnect, error, etc.)
                    await self._recv_task

    async def _recv_loop(self) -> None:
        while True:
            await self._raw_recv()

    async def _subscribe_events(self, event_type: str | None = None) -> int:
        """Subscribe to HA events; returns the subscription ID (the message id you sent)."""
        payload: dict[str, Any] = {"type": "subscribe_events"}
        if event_type is not None:
            payload["event_type"] = event_type  # omit to get all events

        payload["id"] = sub_id = self.get_next_message_id()
        # Use send_and_wait so we see success/error deterministically
        await self.send_and_wait(**payload)
        # HA replies with {'id': <same>, 'type': 'result', 'success': True}
        # We return our own id as the subscription handle for unsubscribe
        return sub_id

    async def handle_start(self) -> None:
        """Handle a start event for the service."""

        if self.status != ResourceStatus.RUNNING:
            await super().handle_start()

    async def _cleanup(self) -> None:
        """Cleanup resources after the WebSocket connection is closed."""

        # Set exceptions for all pending response futures
        for fut in list(self._response_futures.values()):
            if not fut.done():
                fut.set_exception(RetryableConnectionClosedError("WebSocket disconnected"))
        self._response_futures.clear()

        # Try to unsubscribe (best-effort; ignore errors if socket is going away)
        if self._ws and not self._ws.closed and self._subscription_ids:
            for sid in list(self._subscription_ids):
                with suppress(Exception):
                    await self.send_json(type="unsubscribe_events", subscription=sid)
            self._subscription_ids.clear()

        # Stop the recv loop
        if self._recv_task:
            self._recv_task.cancel()
            await asyncio.gather(self._recv_task, return_exceptions=True)
            self._recv_task = None

        # Close the WebSocket
        if self._ws and not self._ws.closed:
            await self._ws.close(
                code=aiohttp.WSCloseCode.GOING_AWAY,
                message=b"Shutting down WebSocket connection",
            )
            self.logger.debug("Closed WebSocket with code %s", aiohttp.WSCloseCode.GOING_AWAY)

        # Close the aiohttp session
        if self._session:
            await self._session.close()
            self.logger.debug("Closed aiohttp session")

    async def send_and_wait(self, **data: Any) -> dict[str, Any]:
        """Send a message and wait for a response.

        Args:
            **data: The data to send as a JSON payload.

        Returns:
            dict[str, Any]: The response data from the WebSocket.

        Raises:
            FailedMessageError: If sending the message fails or times out.
        """
        if "id" not in data:
            data["id"] = msg_id = self.get_next_message_id()
        else:
            msg_id = data["id"]

        fut = asyncio.get_running_loop().create_future()
        self._response_futures[msg_id] = fut
        try:
            await self.send_json(**data)
            return await asyncio.wait_for(fut, timeout=self.timeout_seconds)
        except TimeoutError:
            raise FailedMessageError(f"Response timed out after {self.timeout_seconds}s (data: {data})") from None
        finally:
            self._response_futures.pop(msg_id, None)

    def _respond_if_necessary(self, message: dict) -> None:
        if message.get("type") != "result":
            return

        msg_id = message.get("id")

        if not msg_id:
            self.logger.warning("Received result message without ID: %s", message)
            return

        fut = self._response_futures.get(msg_id)
        if not fut or fut.done():
            return

        if message.get("success"):
            fut.set_result(message.get("result"))

        else:
            err = (message.get("error") or {}).get("message", "Unknown error")
            fut.set_exception(FailedMessageError.from_error_response(err, original_data=message))

    async def send_json(self, **data) -> None:
        """Send a JSON payload over the WebSocket connection, with an incrementing message ID.

        Args:
            **data: The data to send as a JSON payload.

        Raises:
            FailedMessageError: If sending the message fails.
        """
        self.logger.debug("Sending WebSocket message: %s", data)

        if not isinstance(data, dict):
            raise TypeError("Payload must be a dictionary, got %s", type(data).__name__)

        if not self.connected:
            raise ConnectionClosedError("WebSocket connection is not established")

        # this should never be an issue because self.connected checks for this already
        assert self._ws is not None, "WebSocket must be initialized before sending messages"

        if "id" not in data:
            data["id"] = self.get_next_message_id()

        try:
            await self._ws.send_json(data)
        except ClientConnectionResetError:
            self.logger.error("WebSocket connection reset by peer")
            raise

        except Exception as e:
            self.logger.exception("Exception when sending message: %s", data)
            raise FailedMessageError(f"Failed to send message: {data}") from e

    async def authenticate(self) -> None:
        """Authenticate with the Home Assistant WebSocket API."""

        assert self._ws, "WebSocket must be initialized before authenticating"
        token = self.hassette.config.token.get_secret_value()
        truncated_token = self.hassette.config.truncated_token
        ws_url = self.hassette.config.ws_url

        with anyio.fail_after(10):
            msg = await self._ws.receive_json()
            assert msg["type"] == "auth_required"
            await self._ws.send_json({"type": "auth", "access_token": token})
            msg = await self._ws.receive_json()

            # happy path
            if msg["type"] == "auth_ok":
                self.logger.debug("Authenticated successfully with Home Assistant at %s", ws_url)
                return

            if msg["type"] == "auth_invalid":
                self.logger.critical(
                    "Invalid authentication (using token %s) for Home Assistant instance at %s",
                    truncated_token,
                    ws_url,
                )
                raise InvalidAuthError(f"Authentication failed - invalid access token ({truncated_token})")

            raise RuntimeError(f"Unexpected authentication response: {msg}")

    async def _raw_recv(self) -> None:
        """Receive a raw WebSocket frame.

        Raises:
            ConnectionClosedError: If the connection is closed.
        """

        if not self._ws:
            raise RuntimeError("WebSocket connection is not established")

        if self._ws.closed:
            raise RetryableConnectionClosedError("WebSocket connection is closed")

        msg = await self._ws.receive()
        msg_type, raw = msg.type, msg.data

        if msg_type == WSMsgType.TEXT:
            try:
                data = json.loads(raw) if raw else {}
            except json.JSONDecodeError:
                self.logger.exception("Invalid JSON received: %s", raw)
                return

            await self._dispatch(data)
            return

        if msg_type == WSMsgType.BINARY:
            self.logger.warning("Received binary message, which is not expected: %r", raw)
            return

        if msg_type in {WSMsgType.CLOSE, WSMsgType.CLOSED}:
            raise RetryableConnectionClosedError(f"WebSocket closed by peer ({msg_type!r})")

        # took a while to track this one down - we need to cancel if we get told that the connection is closing
        if msg_type == WSMsgType.CLOSING:
            self.logger.debug("WebSocket is closing - exiting receive loop")
            raise RetryableConnectionClosedError("WebSocket is closing")

        self.logger.warning("Received unexpected message type: %r", msg_type)

    async def _dispatch(self, data: dict[str, Any]) -> None:
        try:
            match data.get("type"):
                case "event":
                    await self._dispatch_hass_event(cast("HassEventEnvelopeDict", data))
                case "result":
                    self._respond_if_necessary(data)
                case other:
                    self.logger.debug("Ignoring unknown message type: %s", other)
        except Exception:
            self.logger.exception("Failed to dispatch message: %s", data)

    async def _dispatch_hass_event(self, data: HassEventEnvelopeDict) -> None:
        """Dispatch a Home Assistant event to the event bus."""
        event = create_event_from_hass(data)
        await self.hassette.send_event(event.topic, event)

    async def _send_connection_lost_event(self, error: str) -> None:
        """Send a connection lost event to the event bus."""
        event = create_websocket_status_event(connected=False, error=error)
        await self.hassette.send_event(event.topic, event)



from typing import Literal

UNIT_OF_MEASUREMENT = Literal[
    "mVA",
    "VA",
    "mW",
    "W",
    "kW",
    "MW",
    "GW",
    "TW",
    "BTU/h",
    "mvar",
    "var",
    "kvar",
    "J",
    "kJ",
    "MJ",
    "GJ",
    "mWh",
    "Wh",
    "kWh",
    "MWh",
    "GWh",
    "TWh",
    "cal",
    "kcal",
    "Mcal",
    "Gcal",
    "varh",
    "kvarh",
    "kWh/100km",
    "Wh/km",
    "mi/kWh",
    "km/kWh",
    "mA",
    "A",
    "µV",
    "mV",
    "V",
    "kV",
    "MV",
    "°C",
    "°F",
    "K",
    "μs",
    "ms",
    "s",
    "min",
    "h",
    "d",
    "w",
    "m",
    "y",
    "mm",
    "cm",
    "m",
    "km",
    "in",
    "ft",
    "yd",
    "mi",
    "nmi",
    "Hz",
    "kHz",
    "MHz",
    "GHz",
    "Pa",
    "hPa",
    "kPa",
    "bar",
    "cbar",
    "mbar",
    "mmHg",
    "inHg",
    "psi",
    "dB",
    "dBA",
    "ft³",
    "CCF",
    "m³",
    "L",
    "mL",
    "gal",
    "fl. oz.",
    "m³/h",
    "m³/s",
    "ft³/min",
    "L/h",
    "L/min",
    "L/s",
    "gal/min",
    "mL/s",
    "m²",
    "cm²",
    "km²",
    "mm²",
    "in²",
    "ft²",
    "yd²",
    "mi²",
    "ac",
    "ha",
    "g",
    "kg",
    "mg",
    "µg",
    "oz",
    "lb",
    "st",
    "S/cm",
    "µS/cm",
    "mS/cm",
    "W/m²",
    "BTU/(h⋅ft²)",
    "in/d",
    "in/h",
    "mm/d",
    "mm/h",
    "in",
    "mm",
    "cm",
    "mg/dL",
    "mmol/L",
    "Beaufort",
    "ft/s",
    "in/s",
    "m/s",
    "km/h",
    "kn",
    "mph",
    "mm/s",
    "bit",
    "kbit",
    "Mbit",
    "Gbit",
    "B",
    "kB",
    "MB",
    "GB",
    "TB",
    "PB",
    "EB",
    "ZB",
    "YB",
    "KiB",
    "MiB",
    "GiB",
    "TiB",
    "PiB",
    "EiB",
    "ZiB",
    "YiB",
    "bit/s",
    "kbit/s",
    "Mbit/s",
    "Gbit/s",
    "B/s",
    "kB/s",
    "MB/s",
    "GB/s",
    "KiB/s",
    "MiB/s",
    "GiB/s",
    "°",
    "$",
    "€",
    "¢",
    "dB",
    "dBm",
    "g/m³",
    "µg/m³",
    "mg/m³",
    "μg/ft³",
    "p/m³",
    "ppm",
    "ppb",
    "lx",
    "UV index",
    "%",
    "rpm",
]

DEVICE_CLASS = Literal[
    "date",
    "enum",
    "timestamp",
    "absolute_humidity",
    "apparent_power",
    "aqi",
    "area",
    "atmospheric_pressure",
    "battery",
    "blood_glucose_concentration",
    "carbon_monoxide",
    "carbon_dioxide",
    "conductivity",
    "current",
    "data_rate",
    "data_size",
    "distance",
    "duration",
    "energy",
    "energy_distance",
    "energy_storage",
    "frequency",
    "gas",
    "humidity",
    "illuminance",
    "irradiance",
    "moisture",
    "monetary",
    "nitrogen_dioxide",
    "nitrogen_monoxide",
    "nitrous_oxide",
    "ozone",
    "ph",
    "pm1",
    "pm10",
    "pm25",
    "power_factor",
    "power",
    "precipitation",
    "precipitation_intensity",
    "pressure",
    "reactive_energy",
    "reactive_power",
    "signal_strength",
    "sound_pressure",
    "speed",
    "sulphur_dioxide",
    "temperature",
    "volatile_organic_compounds",
    "volatile_organic_compounds_parts",
    "voltage",
    "volume",
    "volume_storage",
    "volume_flow_rate",
    "water",
    "weight",
    "wind_direction",
    "wind_speed",
]

STATE_CLASS = Literal["measurement", "measurement_angle", "total", "total_increasing"]



from typing import Literal

from pydantic import Field

from .base import AttributesBase, InstantBaseState


class EventState(InstantBaseState):
    class Attributes(AttributesBase):
        event_types: list[str] | None = Field(default=None)
        event_type: str | None = Field(default=None)
        button: str | None = Field(default=None)

    domain: Literal["event"]

    attributes: Attributes



import logging
import typing
from collections.abc import Iterable, Mapping
from contextlib import AsyncExitStack
from datetime import date, datetime
from enum import StrEnum
from logging import getLogger
from typing import Any

import aiohttp
import orjson
from tenacity import before_sleep_log, retry, retry_if_not_exception_type, stop_after_attempt, wait_exponential_jitter
from whenever import Date, Instant, PlainDateTime, ZonedDateTime

from hassette.core.classes import Resource
from hassette.core.events import HassContext, HassStateDict
from hassette.exceptions import ConnectionClosedError, EntityNotFoundError, InvalidAuthError
from hassette.models.entities import BaseEntity, EntityT
from hassette.models.history import HistoryEntry, normalize_history
from hassette.models.states import BaseState, StateT, StateUnion, StateValueT, try_convert_state

if typing.TYPE_CHECKING:
    from hassette.core.core import Hassette
    from hassette.core.websocket import _Websocket

LOGGER = getLogger(__name__)


class _Api(Resource):
    def __init__(self, hassette: "Hassette"):
        super().__init__(hassette)

        self.started = False
        self._stack = AsyncExitStack()

    async def initialize(self):
        """
        Start the API service.
        """
        await super().initialize()

        if self.started:
            return self

        self.logger.debug("Starting '%s' service", self.class_name)

        await self._stack.__aenter__()
        self._session = await self._stack.enter_async_context(
            aiohttp.ClientSession(headers=self._headers, base_url=self._rest_url)
        )

        self.started = True
        return self

    async def shutdown(self, *args, **kwargs) -> None:
        self.started = False
        await self._stack.aclose()
        await super().shutdown()

    @property
    def _headers(self) -> dict[str, str]:
        """Get the headers for this API instance."""
        return self.hassette.config.headers

    @property
    def _rest_url(self) -> str:
        """Get the REST URL for this API instance."""
        return self.hassette.config.rest_url

    @property
    def _ws_conn(self) -> "_Websocket":
        """Get the WebSocket connection for this API instance."""
        return self.hassette._websocket

    @retry(
        retry=retry_if_not_exception_type(
            (EntityNotFoundError, InvalidAuthError, RuntimeError, ConnectionClosedError, TypeError, AttributeError)
        ),
        wait=wait_exponential_jitter(),
        stop=stop_after_attempt(5),
        before_sleep=before_sleep_log(LOGGER, logging.WARNING),
        reraise=True,
    )
    async def _rest_request(
        self,
        method: str,
        url: str,
        params: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
        suppress_error_message: bool = False,
        **kwargs,
    ) -> aiohttp.ClientResponse:
        """Make a REST request to the Home Assistant API."""

        if self._session is None:
            raise RuntimeError("Client session is not connected")

        params = clean_kwargs(**(params or {}))
        str_data = orjson_dump(data or {})

        request_kwargs = {}

        if str_data:
            request_kwargs["data"] = str_data
            request_kwargs["headers"] = {"Content-Type": "application/json"}

        if params:
            request_kwargs["params"] = params

        try:
            response = await self._session.request(method, url, **request_kwargs, **kwargs)
            self.logger.debug("Making %s request to %s with data %s", method, response.real_url, str_data)
            response.raise_for_status()

            return response
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                if not suppress_error_message:
                    self.logger.error("Error occurred while making %s request to %s: %s", method, url, e, stacklevel=2)

                raise EntityNotFoundError(f"Entity not found: {url}") from None
            raise

        except aiohttp.ClientError as e:
            if not suppress_error_message:
                self.logger.error("Error occurred while making %s request to %s: %s", method, url, e, stacklevel=2)

            raise

    async def _get_history_raw(
        self,
        entity_id: str,
        start_time: PlainDateTime | Date | datetime | date | str,
        end_time: PlainDateTime | Date | datetime | date | str | None = None,
        significant_changes_only: bool = False,
        minimal_response: bool = False,
        no_attributes: bool = False,
    ) -> list[list[dict[str, Any]]]:
        """Get the history of a specific entity."""

        url = f"history/period/{start_time}"

        params = {
            "filter_entity_id": entity_id,
            "end_time": end_time,
            "significant_changes_only": significant_changes_only,
            "minimal_response": minimal_response,
            "no_attributes": no_attributes,
        }
        # having parameters like `minimal_response` in the parameters changes the response format
        # regardless of whether they are set to True or False
        # so we remove them if they are False
        params = {k: v for k, v in params.items() if v is not False}

        response = await self._rest_request("GET", url, params=params)

        entries = await response.json()

        normalized = normalize_history(entries)

        return normalized


class Api(Resource):
    """API service for interacting with Home Assistant.

    This service provides methods to interact with the Home Assistant API, including making REST requests,
    managing WebSocket connections, and handling entity states.
    """

    def __init__(self, hassette: "Hassette", _api: _Api) -> None:
        super().__init__(hassette)
        self._api = _api
        self.sync = ApiSyncFacade(self)

    async def ws_send_and_wait(self, **data: Any) -> Any:
        """Send a WebSocket message and wait for a response."""
        return await self._api._ws_conn.send_and_wait(**data)

    async def ws_send_json(self, **data: Any) -> None:
        """Send a WebSocket message without waiting for a response."""
        await self._api._ws_conn.send_json(**data)

    async def rest_request(
        self,
        method: str,
        url: str,
        params: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
        suppress_error_message: bool = False,
        **kwargs,
    ) -> aiohttp.ClientResponse:
        """Make a REST request to the Home Assistant API.

        Args:
            method (str): The HTTP method to use (e.g., "GET", "POST").
            url (str): The URL endpoint for the request.
            params (dict[str, Any], optional): Query parameters for the request.
            data (dict[str, Any], optional): JSON payload for the request.
            suppress_error_message (bool, optional): Whether to suppress error messages.

        Returns:
            aiohttp.ClientResponse: The response from the API.
        """
        return await self._api._rest_request(
            method, url, params=params, data=data, suppress_error_message=suppress_error_message, **kwargs
        )

    async def get_rest_request(
        self, url: str, params: dict[str, Any] | None = None, **kwargs
    ) -> aiohttp.ClientResponse:
        """Make a GET request to the Home Assistant API.

        Args:
            url (str): The URL endpoint for the request.
            params (dict[str, Any], optional): Query parameters for the request.
            kwargs: Additional keyword arguments to pass to the request.

        Returns:
            aiohttp.ClientResponse: The response from the API.
        """
        return await self.rest_request("GET", url, params=params, **kwargs)

    async def post_rest_request(self, url: str, data: dict[str, Any] | None = None, **kwargs) -> aiohttp.ClientResponse:
        """Make a POST request to the Home Assistant API.

        Args:
            url (str): The URL endpoint for the request.
            data (dict[str, Any], optional): JSON payload for the request.
            kwargs: Additional keyword arguments to pass to the request.

        Returns:
            aiohttp.ClientResponse: The response from the API.
        """
        return await self.rest_request("POST", url, data=data, **kwargs)

    async def delete_rest_request(self, url: str, **kwargs) -> aiohttp.ClientResponse:
        """Make a DELETE request to the Home Assistant API.

        Args:
            url (str): The URL endpoint for the request.
            kwargs: Additional keyword arguments to pass to the request.

        Returns:
            aiohttp.ClientResponse: The response from the API.
        """
        return await self.rest_request("DELETE", url, **kwargs)

    async def get_states_raw(self) -> list[HassStateDict]:
        """Get all entities in Home Assistant as raw dictionaries.

        Returns:
            list[HassStateDict]: A list of states as dictionaries.
        """
        val: list[HassStateDict] = await self.ws_send_and_wait(type="get_states")  # type: ignore
        assert isinstance(val, list), "Expected a list of states"
        return val

    async def get_states(self) -> list[StateUnion]:
        """Get all entities in Home Assistant.

        Returns:
            list[StateUnion]: A list of states, either as dictionaries or converted to state objects.
        """
        val = await self.get_states_raw()

        self.logger.debug("Converting states to specific state types")
        return list(filter(bool, [try_convert_state(state) for state in val]))

    async def get_config(self) -> dict[str, Any]:
        """
        Get the Home Assistant configuration.

        Returns:
            dict: The configuration data.
        """
        val = await self.ws_send_and_wait(type="get_config")
        assert isinstance(val, dict), "Expected a dictionary of configuration data"
        return val

    async def get_services(self) -> dict[str, Any]:
        """
        Get the available services in Home Assistant.

        Returns:
            dict: The services data.
        """
        val = await self.ws_send_and_wait(type="get_services")
        assert isinstance(val, dict), "Expected a dictionary of services"
        return val

    async def get_panels(self) -> dict[str, Any]:
        """
        Get the available panels in Home Assistant.

        Returns:
            dict: The panels data.
        """
        val = await self.ws_send_and_wait(type="get_panels")
        assert isinstance(val, dict), "Expected a dictionary of panels"
        return val

    async def fire_event(
        self,
        event_type: str,
        event_data: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """
        Fire a custom event in Home Assistant.

        Args:
            event_type (str): The type of the event to fire (e.g., "custom_event").
            event_data (dict[str, Any], optional): Additional data to include with the event.

        Returns:
            dict: The response from Home Assistant.
        """
        event_data = event_data or {}

        data = {"type": "fire_event", "event_type": event_type, "event_data": event_data}
        if not event_data:
            data.pop("event_data")

        return await self.ws_send_and_wait(**data)

    async def call_service(
        self,
        domain: str,
        service: str,
        target: dict[str, str] | dict[str, list[str]] | None = None,
        return_response: bool | None = None,
        **data,
    ) -> HassContext | None:
        """
        Call a Home Assistant service.

        Args:
            domain (str): The domain of the service (e.g., "light").
            service (str): The name of the service to call (e.g., "turn_on").
            target (dict[str, str], optional): Target entity IDs or areas.
            **kwargs: Additional data to send with the service call.

        Returns:
            HassContext | None: The response from Home Assistant if return_response is True. Otherwise, returns None.
        """
        payload = {
            "type": "call_service",
            "domain": domain,
            "service": service,
            "target": target,
        }

        payload = {k: v for k, v in payload.items() if v is not None}

        if data:
            data = {k: v for k, v in data.items() if v is not None}
            self.logger.debug("Adding extra data to service call: %s", data)
            payload |= {"service_data": data}

        if return_response:
            resp = await self.ws_send_and_wait(**payload)
            return HassContext(**resp.get("context", {}))

        await self.ws_send_json(**payload)
        return None

    async def turn_on(self, entity_id: str | StrEnum, domain: str = "homeassistant", **data):
        """
        Turn on a specific entity in Home Assistant.

        Args:
            entity_id (str): The ID of the entity to turn on (e.g., "light.office").
            domain (str): The domain of the entity (default: "homeassistant").

        Returns:
            HassContext: The response context from Home Assistant.
        """
        entity_id = str(entity_id)

        self.logger.debug("Turning on entity %s", entity_id)
        return await self.call_service(
            domain=domain,
            service="turn_on",
            target={"entity_id": entity_id},
            return_response=True,
            **data,
        )

    async def turn_off(self, entity_id: str, domain: str = "homeassistant"):
        """
        Turn off a specific entity in Home Assistant.

        Args:
            entity_id (str): The ID of the entity to turn off (e.g., "light.office").
            domain (str): The domain of the entity (default: "homeassistant").

        Returns:
            HassContext: The response context from Home Assistant.
        """
        self.logger.debug("Turning off entity %s", entity_id)
        return await self.call_service(
            domain=domain,
            service="turn_off",
            target={"entity_id": entity_id},
            return_response=True,
        )

    async def toggle_service(self, entity_id: str, domain: str = "homeassistant"):
        """
        Toggle a specific entity in Home Assistant.

        Args:
            entity_id (str): The ID of the entity to toggle (e.g., "light.office").
            domain (str): The domain of the entity (default: "homeassistant").

        Returns:
            HassContext: The response context from Home Assistant.
        """
        self.logger.debug("Toggling entity %s", entity_id)
        return await self.call_service(
            domain=domain,
            service="toggle",
            target={"entity_id": entity_id},
            return_response=True,
        )

    async def get_state_raw(self, entity_id: str) -> HassStateDict:
        """Get the state of a specific entity.

        Args:
            entity_id (str): The ID of the entity to get the state for.

        Returns:
            HassStateDict: The state of the entity as raw data.
        """

        url = f"states/{entity_id}"
        response = await self.get_rest_request(url)
        return await response.json()

    async def entity_exists(self, entity_id: str) -> bool:
        """Check if a specific entity exists.

        Args:
            entity_id (str): The ID of the entity to check.

        Returns:
            bool: True if the entity exists, False otherwise.
        """

        try:
            url = f"states/{entity_id}"
            response = await self.rest_request("GET", url, suppress_error_message=True)
            await response.json()
            return True
        except EntityNotFoundError:
            return False

    async def get_entity(self, entity_id: str, model: type[EntityT]) -> EntityT:
        """Get an entity object for a specific entity.

        Args:
            entity_id (str): The ID of the entity to get.
            model (type[EntityT]): The model class to use for the entity.

        Returns:
            EntityT: The entity object.

        Note:
            This is not the same as calling get_state: get_state returns a BaseState subclass.
            This call returns an EntityState subclass, which wraps the state object and provides
            api methods for interacting with the entity.

        """
        if not issubclass(model, BaseEntity):  # runtime check
            raise TypeError(f"Model {model!r} is not a valid BaseEntity subclass")

        raw = await self.get_state_raw(entity_id)

        return model.model_validate({"state": raw})

    async def get_entity_or_none(self, entity_id: str, model: type[EntityT]) -> EntityT | None:
        """Get an entity object for a specific entity, or None if it does not exist.

        Args:
            entity_id (str): The ID of the entity to get.
            model (type[EntityT]): The model class to use for the entity.

        Returns:
            EntityT | None: The entity object, or None if it does not exist.
        """
        try:
            return await self.get_entity(entity_id, model)
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                return None
            raise

    async def get_state(self, entity_id: str, model: type[StateT]) -> StateT:
        """Get the state of a specific entity.

        Args:
            entity_id (str): The ID of the entity to get the state for.
            model (type[StateT]): The model type to convert the state to.

        Returns:
            StateT: The state of the entity converted to the specified model type.
        """

        if not issubclass(model, BaseState):  # runtime check
            raise TypeError(f"Model {model!r} is not a valid StateType subclass")

        raw = await self.get_state_raw(entity_id)

        return model.model_validate(raw)

    async def get_state_value(self, entity_id: str) -> str:
        """Get the state of a specific entity without converting it to a state object.

        Args:
            entity_id (str): The ID of the entity to get the state for.

        Returns:
            str: The state of the entity as raw data.

        Note:
            While most default methods in this library work with state objects for
            strong typing, this method is designed to return the raw state value,
            as it is likely overkill to convert it to a state object for simple state value retrieval.
        """

        entity = await self.get_state_raw(entity_id)
        state = entity.get("state")
        if not isinstance(state, str):
            self.logger.info(
                "Entity %s state is not a string (%s), return type annotation should be updated",
                entity_id,
                type(state).__name__,
            )

        return state  # pyright: ignore[reportReturnType]

    async def get_state_value_typed(self, entity_id: str, model: type[BaseState[StateValueT]]) -> StateValueT:
        """Get the state of a specific entity as a converted state object.

        Args:
            entity_id (str): The ID of the entity to get the state for.
            model (type[BaseState[StateValueT]]): The model type to convert the state to.

        Returns:
            StateValueT: The state of the entity converted to the specified model type.

        Raises:
            TypeError: If the model is not a valid StateType subclass.

        Note:
            Instead of the default way of calling `get_state` involving a type, we assume that the
            average user only needs the raw value of the state value, without type safety.
        """

        state = await self.get_state(entity_id, model)
        return state.value

    async def get_attribute(self, entity_id: str, attribute: str) -> Any | None:
        """Get a specific attribute of an entity.

        Args:
            entity_id (str): The ID of the entity to get the attribute for.
            attribute (str): The name of the attribute to retrieve.

        Returns:
            Any: The value of the specified attribute, or None if it does not exist.
        """

        entity = await self.get_state_raw(entity_id)
        return (entity.get("attributes", {}) or {}).get(attribute)

    async def get_history(
        self,
        entity_id: str,
        start_time: PlainDateTime | Date | datetime | date | str,
        end_time: PlainDateTime | Date | datetime | date | str | None = None,
        significant_changes_only: bool = False,
        minimal_response: bool = False,
        no_attributes: bool = False,
    ) -> list[HistoryEntry]:
        """Get the history of a specific entity.

        Args:
            entity_id (str): The ID of the entity to get the history for.
            start_time (PlainDateTime | Date | datetime | date | str):
                The start time for the history range.
            end_time (PlainDateTime | Date | datetime | date | str | None, optional):
                The end time for the history range.
            significant_changes_only (bool, optional): Whether to only include significant changes.
            minimal_response (bool, optional): Whether to request a minimal response.
            no_attributes (bool, optional): Whether to exclude attributes from the response.

        Returns:
            list[HistoryEntry]: A list of history entries for the specified entity.
        """
        if "," in entity_id:
            raise ValueError("Entity ID should not contain commas. Use `get_histories` for multiple entities.")

        entries = await self._api._get_history_raw(
            entity_id=entity_id,
            start_time=start_time,
            end_time=end_time,
            significant_changes_only=significant_changes_only,
            minimal_response=minimal_response,
            no_attributes=no_attributes,
        )

        if not entries:
            return []

        assert len(entries) == 1, "Expected a single list of history entries"

        converted = [HistoryEntry.model_validate(entry) for entry in entries[0]]

        return converted

    async def get_histories(
        self,
        entity_ids: list[str],
        start_time: PlainDateTime | Date | datetime | date | str,
        end_time: PlainDateTime | Date | datetime | date | str | None = None,
        significant_changes_only: bool = False,
        minimal_response: bool = False,
        no_attributes: bool = False,
    ) -> dict[str, list[HistoryEntry]]:
        """Get the history for multiple entities.

        Args:
            entity_ids (list[str]): The IDs of the entities to get the history for.
            start_time (PlainDateTime | Date | datetime | date | str):
                The start time for the history range.
            end_time (PlainDateTime | Date | datetime | date | str | None, optional):
                The end time for the history range.
            significant_changes_only (bool, optional): Whether to only include significant changes.
            minimal_response (bool, optional): Whether to request a minimal response.
            no_attributes (bool, optional): Whether to exclude attributes from the response.

        Returns:
            dict[str, list[HistoryEntry]]: A dictionary mapping entity IDs to their respective history entries.
        """
        entity_id = ",".join(entity_ids)

        entries = await self._api._get_history_raw(
            entity_id=entity_id,
            start_time=start_time,
            end_time=end_time,
            significant_changes_only=significant_changes_only,
            minimal_response=minimal_response,
            no_attributes=no_attributes,
        )

        if not entries:
            return {}

        converted = {}
        for history_list in entries:
            converted[history_list[0]["entity_id"]] = [HistoryEntry.model_validate(entry) for entry in history_list]

        return converted

    async def get_logbook(
        self,
        entity_id: str,
        start_time: PlainDateTime | Date | datetime | date | str,
        end_time: PlainDateTime | Date | datetime | date | str,
    ) -> list[dict]:
        """Get the logbook entries for a specific entity.

        Args:
            entity_id (str): The ID of the entity to get the logbook entries for.
            start_time (PlainDateTime | Date | datetime | date | str): The start time for the logbook range.
            end_time (PlainDateTime | Date | datetime | date | str): The end time for the logbook range.

        Returns:
            list[dict]: A list of logbook entries for the specified entity.
        """

        url = f"logbook/{start_time}"
        params = {"entity": entity_id, "end_time": end_time}

        response = await self.get_rest_request(url, params=params)

        return await response.json()

    async def set_state(
        self,
        entity_id: str | StrEnum,
        state: str,
        attributes: dict[str, Any] | None = None,
    ) -> dict:
        """Set the state of a specific entity.

        Args:
            entity_id (str | StrEnum): The ID of the entity to set the state for.
            state (str): The new state value to set.
            attributes (dict[str, Any], optional): Additional attributes to set for the entity.

        Returns:
            dict: The response from Home Assistant after setting the state.
        """

        entity_id = str(entity_id)

        attributes = attributes or {}
        curr_attributes = {}

        if await self.entity_exists(entity_id):
            curr_attributes = (await self.get_state_raw(entity_id)).get("attributes", {}) or {}

        # Merge current attributes with new attributes
        new_attributes = curr_attributes | attributes

        url = f"states/{entity_id}"
        data = {"state": state, "attributes": new_attributes}

        response = await self.post_rest_request(url, data=data)
        return await response.json()

    async def get_camera_image(
        self,
        entity_id: str,
        timestamp: ZonedDateTime | PlainDateTime | Date | datetime | date | str | None = None,
    ) -> bytes:
        """Get the latest camera image for a specific entity.

        Args:
            entity_id (str): The ID of the camera entity to get the image for.
            timestamp (ZonedDateTime | PlainDateTime | Date | datetime | date | str | None, optional):
                The timestamp for the image. If None, the latest image is returned.

        Returns:
            bytes: The camera image data.
        """

        url = f"camera_proxy/{entity_id}"
        params = {}
        if timestamp:
            params["timestamp"] = timestamp

        response = await self.get_rest_request(url, params=params)

        return await response.read()

    async def get_calendars(self) -> list[dict]:
        """Get the list of calendars."""

        url = "calendars"
        response = await self.get_rest_request(url)
        return await response.json()

    async def get_calendar_events(
        self,
        calendar_id: str,
        start_time: PlainDateTime | Date | datetime | date | str,
        end_time: PlainDateTime | Date | datetime | date | str,
    ) -> list[dict]:
        """Get events from a specific calendar.

        Args:
            calendar_id (str): The ID of the calendar to get events from.
            start_time (PlainDateTime | Date | datetime | date | str): The start time for the event range.
            end_time (PlainDateTime | Date | datetime | date | str): The end time for the event range.

        Returns:
            list[dict]: A list of calendar events.
        """

        url = f"calendars/{calendar_id}/events"
        params = {"start": start_time, "end": end_time}

        response = await self.get_rest_request(url, params=params)
        return await response.json()

    async def render_template(
        self,
        template: str,
        variables: dict | None = None,
    ) -> str:
        """Render a template with given variables.

        Args:
            template (str): The template string to render.
            variables (dict, optional): Variables to use in the template.

        Returns:
            str: The rendered template result.
        """

        url = "template"
        data = {"template": template, "variables": variables or {}}

        response = await self.post_rest_request(url, data=data)
        return await response.text()

    async def delete_entity(self, entity_id: str) -> None:
        """Delete a specific entity.

        Args:
            entity_id (str): The ID of the entity to delete.

        Raises:
            RuntimeError: If the deletion fails.
        """

        url = f"states/{entity_id}"

        response = await self.rest_request("DELETE", url)

        if response.status != 204:
            raise RuntimeError(f"Failed to delete entity {entity_id}: {response.status} - {response.reason}")


class ApiSyncFacade(Resource):
    """Synchronous facade for the API service.

    This class provides synchronous methods that wrap the asynchronous methods of the Api class,
    allowing for blocking calls in a synchronous context.

    It is important to note that these methods should not be called from within an existing event loop,
    as they will raise a RuntimeError in such cases. Use the asynchronous methods directly when operating
    within an event loop.
    """

    def __init__(self, api: Api):
        super().__init__(api.hassette)
        self._api = api

    def ws_send_and_wait(self, **data: Any):
        """Send a WebSocket message and wait for a response."""
        return self.hassette.run_sync(self._api.ws_send_and_wait(**data))

    def ws_send_json(self, **data: Any):
        """Send a WebSocket message without waiting for a response."""
        return self.hassette.run_sync(self._api.ws_send_json(**data))

    def rest_request(
        self,
        method: str,
        url: str,
        params: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
        suppress_error_message: bool = False,
        **kwargs,
    ):
        """Make a REST request to the Home Assistant API.

        Args:
            method (str): The HTTP method to use (e.g., "GET", "POST").
            url (str): The URL endpoint for the request.
            params (dict[str, Any], optional): Query parameters for the request.
            data (dict[str, Any], optional): JSON payload for the request.
            suppress_error_message (bool, optional): Whether to suppress error messages.

        Returns:
            aiohttp.ClientResponse: The response from the API.
        """
        return self.hassette.run_sync(
            self._api.rest_request(
                method, url, params=params, data=data, suppress_error_message=suppress_error_message, **kwargs
            )
        )

    def get_rest_request(self, url: str, params: dict[str, Any] | None = None, **kwargs):
        """Make a GET request to the Home Assistant API.

        Args:
            url (str): The URL endpoint for the request.
            params (dict[str, Any], optional): Query parameters for the request.
            kwargs: Additional keyword arguments to pass to the request.

        Returns:
            aiohttp.ClientResponse: The response from the API.
        """
        return self.hassette.run_sync(self._api.get_rest_request(url, params=params, **kwargs))

    def post_rest_request(self, url: str, data: dict[str, Any] | None = None, **kwargs):
        """Make a POST request to the Home Assistant API.

        Args:
            url (str): The URL endpoint for the request.
            data (dict[str, Any], optional): JSON payload for the request.
            kwargs: Additional keyword arguments to pass to the request.

        Returns:
            aiohttp.ClientResponse: The response from the API.
        """
        return self.hassette.run_sync(self._api.post_rest_request(url, data=data, **kwargs))

    def delete_rest_request(self, url: str, **kwargs):
        """Make a DELETE request to the Home Assistant API.

        Args:
            url (str): The URL endpoint for the request.
            kwargs: Additional keyword arguments to pass to the request.

        Returns:
            aiohttp.ClientResponse: The response from the API.
        """
        return self.hassette.run_sync(self._api.delete_rest_request(url, **kwargs))

    def get_states_raw(self):
        """Get all entities in Home Assistant as raw dictionaries.

        Returns:
            list[HassStateDict]: A list of states as dictionaries.
        """
        return self.hassette.run_sync(self._api.get_states_raw())

    def get_states(self):
        """Get all entities in Home Assistant.

        Returns:
            list[StateUnion]: A list of states, either as dictionaries or converted to state objects.
        """
        return self.hassette.run_sync(self._api.get_states())

    def get_config(self):
        """
        Get the Home Assistant configuration.

        Returns:
            dict: The configuration data.
        """
        return self.hassette.run_sync(self._api.get_config())

    def get_services(self):
        """
        Get the available services in Home Assistant.

        Returns:
            dict: The services data.
        """
        return self.hassette.run_sync(self._api.get_services())

    def get_panels(self):
        """
        Get the available panels in Home Assistant.

        Returns:
            dict: The panels data.
        """
        return self.hassette.run_sync(self._api.get_panels())

    def fire_event(
        self,
        event_type: str,
        event_data: dict[str, Any] | None = None,
    ):
        """
        Fire a custom event in Home Assistant.

        Args:
            event_type (str): The type of the event to fire (e.g., "custom_event").
            event_data (dict[str, Any], optional): Additional data to include with the event.

        Returns:
            dict: The response from Home Assistant.
        """
        return self.hassette.run_sync(self._api.fire_event(event_type, event_data))

    def call_service(
        self,
        domain: str,
        service: str,
        target: dict[str, str] | dict[str, list[str]] | None = None,
        return_response: bool = True,
        **data,
    ):
        """
        Call a Home Assistant service.

        Args:
            domain (str): The domain of the service (e.g., "light").
            service (str): The name of the service to call (e.g., "turn_on").
            target (dict[str, str], optional): Target entity IDs or areas.
            **kwargs: Additional data to send with the service call.

        Returns:
            HassContext | None: The response from Home Assistant if return_response is True.
                Otherwise, returns None.
        """
        return self.hassette.run_sync(self._api.call_service(domain, service, target, return_response, **data))

    def turn_on(self, entity_id: str | StrEnum, domain: str = "homeassistant", **data):
        """
        Turn on a specific entity in Home Assistant.

        Args:
            entity_id (str): The ID of the entity to turn on (e.g., "light.office").
            domain (str): The domain of the entity (default: "homeassistant").

        Returns:
            HassContext: The response context from Home Assistant.
        """
        return self.hassette.run_sync(self._api.turn_on(entity_id, domain, **data))

    def turn_off(self, entity_id: str, domain: str = "homeassistant"):
        """
        Turn off a specific entity in Home Assistant.

        Args:
            entity_id (str): The ID of the entity to turn off (e.g., "light.office").
            domain (str): The domain of the entity (default: "homeassistant").

        Returns:
            HassContext: The response context from Home Assistant.
        """
        return self.hassette.run_sync(self._api.turn_off(entity_id, domain))

    def toggle_service(self, entity_id: str, domain: str = "homeassistant"):
        """
        Toggle a specific entity in Home Assistant.

        Args:
            entity_id (str): The ID of the entity to toggle (e.g., "light.office").
            domain (str): The domain of the entity (default: "homeassistant").

        Returns:
            HassContext: The response context from Home Assistant.
        """
        return self.hassette.run_sync(self._api.toggle_service(entity_id, domain))

    def get_state_raw(self, entity_id: str):
        """Get the state of a specific entity.

        Args:
            entity_id (str): The ID of the entity to get the state for.

        Returns:
            HassStateDict: The state of the entity as raw data.
        """
        return self.hassette.run_sync(self._api.get_state_raw(entity_id))

    def entity_exists(self, entity_id: str):
        """Check if a specific entity exists.

        Args:
            entity_id (str): The ID of the entity to check.

        Returns:
            bool: True if the entity exists, False otherwise.
        """

        return self.hassette.run_sync(self._api.entity_exists(entity_id))

    def get_entity(self, entity_id: str, model: type[EntityT]):
        """Get an entity object for a specific entity.

        Args:
            entity_id (str): The ID of the entity to get.
            model (type[EntityT]): The model class to use for the entity.

        Returns:
            EntityT: The entity object.

        Note:
            This is not the same as calling get_state: get_state returns a BaseState subclass.
            This call returns an EntityState subclass, which wraps the state object and provides
            api methods for interacting with the entity.
        """
        return self.hassette.run_sync(self._api.get_entity(entity_id, model))

    def get_entity_or_none(self, entity_id: str, model: type[EntityT]):
        """Get an entity object for a specific entity, or None if it does not exist.

        Args:
            entity_id (str): The ID of the entity to get.
            model (type[EntityT]): The model class to use for the entity.

        Returns:
            EntityT | None: The entity object, or None if it does not exist.
        """
        return self.hassette.run_sync(self._api.get_entity_or_none(entity_id, model))

    def get_state(self, entity_id: str, model: type[StateT]):
        """Get the state of a specific entity.

        Args:
            entity_id (str): The ID of the entity to get the state for.
            model (type[StateT]): The model type to convert the state to.

        Returns:
            StateT: The state of the entity converted to the specified model type.
        """
        return self.hassette.run_sync(self._api.get_state(entity_id, model))

    def get_state_value(self, entity_id: str):
        """Get the state of a specific entity without converting it to a state object.

        Args:
            entity_id (str): The ID of the entity to get the state for.

        Returns:
            str: The state of the entity as raw data.

        Note:
            While most default methods in this library work with state objects for
            strong typing, this method is designed to return the raw state value,
            as it is likely overkill to convert it to a state object for simple state value retrieval.
        """
        return self.hassette.run_sync(self._api.get_state_value(entity_id))

    def get_state_value_typed(self, entity_id: str, model: type[BaseState[StateValueT]]):
        """Get the state of a specific entity as a converted state object.

        Args:
            entity_id (str): The ID of the entity to get the state for.
            model (type[BaseState[StateValueT]]): The model type to convert the state to.

        Returns:
            StateValueT: The state of the entity converted to the specified model type.

        Raises:
            TypeError: If the model is not a valid StateType subclass.

        Note:
            Instead of the default way of calling `get_state` involving a type, we assume that the
            average user only needs the raw value of the state value, without type safety.
        """

        return self.hassette.run_sync(self._api.get_state_value_typed(entity_id, model))

    def get_attribute(self, entity_id: str, attribute: str):
        """Get a specific attribute of an entity.

        Args:
            entity_id (str): The ID of the entity to get the attribute for.
            attribute (str): The name of the attribute to retrieve.

        Returns:
            Any: The value of the specified attribute, or None if it does not exist.
        """

        return self.hassette.run_sync(self._api.get_attribute(entity_id, attribute))

    def get_history(
        self,
        entity_id: str,
        start_time: PlainDateTime | Date | datetime | date | str,
        end_time: PlainDateTime | Date | datetime | date | str | None = None,
        significant_changes_only: bool = False,
        minimal_response: bool = False,
        no_attributes: bool = False,
    ):
        """Get the history of a specific entity.

        Args:
            entity_id (str): The ID of the entity to get the history for.
            start_time (PlainDateTime | Date | datetime | date | str):
                The start time for the history range.
            end_time (PlainDateTime | Date | datetime | date | str | None, optional):
                The end time for the history range.
            significant_changes_only (bool, optional): Whether to only include significant changes.
            minimal_response (bool, optional): Whether to request a minimal response.
            no_attributes (bool, optional): Whether to exclude attributes from the response.

        Returns:
            list[HistoryEntry]: A list of history entries for the specified entity.
        """
        return self.hassette.run_sync(
            self._api.get_history(
                entity_id=entity_id,
                start_time=start_time,
                end_time=end_time,
                significant_changes_only=significant_changes_only,
                minimal_response=minimal_response,
                no_attributes=no_attributes,
            )
        )

    def get_histories(
        self,
        entity_ids: list[str],
        start_time: PlainDateTime | Date | datetime | date | str,
        end_time: PlainDateTime | Date | datetime | date | str | None = None,
        significant_changes_only: bool = False,
        minimal_response: bool = False,
        no_attributes: bool = False,
    ):
        """Get the history for multiple entities.

        Args:
            entity_ids (list[str]): The IDs of the entities to get the history for.
            start_time (PlainDateTime | Date | datetime | date | str):
                The start time for the history range.
            end_time (PlainDateTime | Date | datetime | date | str | None, optional):
                The end time for the history range.
            significant_changes_only (bool, optional): Whether to only include significant changes.
            minimal_response (bool, optional): Whether to request a minimal response.
            no_attributes (bool, optional): Whether to exclude attributes from the response.

        Returns:
            dict[str, list[HistoryEntry]]: A dictionary mapping entity IDs to their respective history entries.
        """
        return self.hassette.run_sync(
            self._api.get_histories(
                entity_ids=entity_ids,
                start_time=start_time,
                end_time=end_time,
                significant_changes_only=significant_changes_only,
                minimal_response=minimal_response,
                no_attributes=no_attributes,
            )
        )

    def get_logbook(
        self,
        entity_id: str,
        start_time: PlainDateTime | Date | datetime | date | str,
        end_time: PlainDateTime | Date | datetime | date | str,
    ):
        """Get the logbook entries for a specific entity.

        Args:
            entity_id (str): The ID of the entity to get the logbook entries for.
            start_time (PlainDateTime | Date | datetime | date | str): The start time for the logbook range.
            end_time (PlainDateTime | Date | datetime | date | str): The end time for the logbook range.

        Returns:
            list[dict]: A list of logbook entries for the specified entity.
        """

        return self.hassette.run_sync(self._api.get_logbook(entity_id, start_time, end_time))

    def set_state(
        self,
        entity_id: str | StrEnum,
        state: str,
        attributes: dict[str, Any] | None = None,
    ):
        """Set the state of a specific entity.

        Args:
            entity_id (str | StrEnum): The ID of the entity to set the state for.
            state (str): The new state value to set.
            attributes (dict[str, Any], optional): Additional attributes to set for the entity.

        Returns:
            dict: The response from Home Assistant after setting the state.
        """

        return self.hassette.run_sync(self._api.set_state(entity_id, state, attributes))

    def get_camera_image(
        self,
        entity_id: str,
        timestamp: ZonedDateTime | PlainDateTime | Date | datetime | date | str | None = None,
    ):
        """Get the latest camera image for a specific entity.

        Args:
            entity_id (str): The ID of the camera entity to get the image for.
            timestamp (ZonedDateTime | PlainDateTime | Date | datetime | date | str | None, optional):
                The timestamp for the image. If None, the latest image is returned.

        Returns:
            bytes: The camera image data.
        """

        return self.hassette.run_sync(self._api.get_camera_image(entity_id, timestamp))

    def get_calendars(self):
        """Get the list of calendars.

        Returns:
            list[dict]: The calendars configured in Home Assistant.
        """

        return self.hassette.run_sync(self._api.get_calendars())

    def get_calendar_events(
        self,
        calendar_id: str,
        start_time: PlainDateTime | Date | datetime | date | str,
        end_time: PlainDateTime | Date | datetime | date | str,
    ):
        """Get events from a specific calendar.

        Args:
            calendar_id (str): The ID of the calendar to get events from.
            start_time (PlainDateTime | Date | datetime | date | str): The start time for the event range.
            end_time (PlainDateTime | Date | datetime | date | str): The end time for the event range.

        Returns:
            list[dict]: A list of calendar events.
        """

        return self.hassette.run_sync(
            self._api.get_calendar_events(
                calendar_id=calendar_id,
                start_time=start_time,
                end_time=end_time,
            )
        )

    def render_template(
        self,
        template: str,
        variables: dict | None = None,
    ):
        """Render a template with given variables.

        Args:
            template (str): The template string to render.
            variables (dict, optional): Variables to use in the template.

        Returns:
            str: The rendered template result.
        """
        return self.hassette.run_sync(self._api.render_template(template, variables))

    def delete_entity(self, entity_id: str):
        """Delete a specific entity.

        Args:
            entity_id (str): The ID of the entity to delete.

        Raises:
            RuntimeError: If the deletion fails.
        """

        self.hassette.run_sync(self._api.delete_entity(entity_id))


def orjson_dump(data: Any) -> str:
    return orjson.dumps(data, default=str).decode("utf-8")


def clean_kwargs(**kwargs: Any) -> dict[str, Any]:
    """Converts values to strings where needed and removes keys with None values."""

    def clean_value(val: Any) -> Any:
        if val is None:
            return None

        if isinstance(val, bool):
            return str(val).lower()

        if isinstance(val, (PlainDateTime | ZonedDateTime | Instant | Date)):
            return val.format_common_iso()

        if isinstance(val, (int | float | str)):
            if isinstance(val, str) and not val.strip():
                return None
            return val

        if isinstance(val, datetime):
            return val.isoformat()

        if isinstance(val, Mapping):
            return {k: clean_value(v) for k, v in val.items() if v is not None}

        if isinstance(val, Iterable) and not isinstance(val, str | bytes):
            return [clean_value(v) for v in val if v is not None]

        return str(val)

    return {k: cleaned for k, v in kwargs.items() if (cleaned := clean_value(v)) is not None}



from typing import Literal

from pydantic import Field

from hassette.const.sensor import DEVICE_CLASS, STATE_CLASS, UNIT_OF_MEASUREMENT

from .base import AttributesBase, StringBaseState


class SensorAttributes(AttributesBase):
    device_class: DEVICE_CLASS | str | None = Field(default=None)

    state_class: STATE_CLASS | None | str = Field(default=None)
    unit_of_measurement: UNIT_OF_MEASUREMENT | str | None = Field(default=None)

    options: list[str] | None = Field(default=None)


class SensorState(StringBaseState):
    domain: Literal["sensor"]
    attributes: SensorAttributes



from typing import Any, Literal

from pydantic import Field
from whenever import Instant, SystemDateTime

from .base import AttributesBase, BoolBaseState, DateTimeBaseState, NumericBaseState, StringBaseState


class InputAttributesBase(AttributesBase):
    """Base attributes class for all input states."""

    editable: bool | None = Field(default=None)


class InputBooleanState(BoolBaseState):
    domain: Literal["input_boolean"]

    attributes: InputAttributesBase


class InputButtonState(DateTimeBaseState):
    domain: Literal["input_button"]

    attributes: InputAttributesBase


class InputDatetimeState(DateTimeBaseState):
    class Attributes(InputAttributesBase):
        has_date: bool | None = Field(default=None)
        has_time: bool | None = Field(default=None)
        year: int | None = Field(default=None)
        month: int | None = Field(default=None)
        day: int | None = Field(default=None)
        hour: int | None = Field(default=None)
        minute: int | None = Field(default=None)
        second: int | None = Field(default=None)
        timestamp: float | None = Field(default=None)

        @property
        def timestamp_as_instant(self) -> Instant | None:
            if self.timestamp is None:
                return None
            return Instant.from_timestamp(self.timestamp)

        @property
        def timestamp_as_system_datetime(self) -> SystemDateTime | None:
            if self.timestamp is None:
                return None
            return SystemDateTime.from_timestamp(self.timestamp)

    domain: Literal["input_datetime"]

    attributes: Attributes


class InputNumberState(NumericBaseState):
    class Attributes(InputAttributesBase):
        max: float | None = Field(default=None)
        initial: float | None = Field(default=None)
        step: int | float | None = Field(default=None)
        mode: str | None = Field(default=None)
        min: int | float | None = Field(default=None)

    domain: Literal["input_number"]

    attributes: Attributes


class InputSelectState(StringBaseState):
    class Attributes(InputAttributesBase):
        options: list[str] = Field(default_factory=list)

    domain: Literal["input_select"]

    attributes: Attributes


class InputTextState(StringBaseState):
    class Attributes(InputAttributesBase):
        min: int | float | None = Field(default=None)
        max: int | float | None = Field(default=None)
        pattern: Any | None = Field(default=None)
        mode: str | None = Field(default=None)

    domain: Literal["input_text"]

    attributes: Attributes



import logging
import sys
import threading
from contextlib import suppress
from typing import Literal

import coloredlogs

FORMAT_DATE = "%Y-%m-%d"
FORMAT_TIME = "%H:%M:%S"
FORMAT_DATETIME = f"{FORMAT_DATE} {FORMAT_TIME}"
FMT = "%(asctime)s %(levelname)s %(name)s.%(funcName)s:%(lineno)d ─ %(message)s"

# TODO: remove coloredlogs and roll our own? or use colorlogs
# coloredlogs is unmaintained and parts of it are broken on Python 3.13+


def enable_logging(
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
) -> None:
    """Set up the logging"""

    logger = logging.getLogger("hassette")

    coloredlogs.install(level=log_level, logger=logger, fmt=FMT, datefmt=FORMAT_DATETIME)

    # Move the coloredlogs handler to hassette logger, not root
    # TODO: move off coloredlogs, this stuff is too buggy
    with suppress(IndexError):
        logger.addHandler(logging.getLogger().handlers.pop(0))

    # set hassette log level back to declared level
    logger.setLevel(log_level)

    # Capture warnings.warn(...) and friends messages in logs.
    # The standard destination for them is stderr, which may end up unnoticed.
    # This way they're where other messages are, and can be filtered as usual.
    logging.captureWarnings(True)

    # Suppress overly verbose logs from libraries that aren't helpful
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("aiohttp.access").setLevel(logging.WARNING)
    logging.getLogger("httpx").setLevel(logging.WARNING)

    sys.excepthook = lambda *args: logging.getLogger().exception("Uncaught exception", exc_info=args)
    threading.excepthook = lambda args: logging.getLogger().exception(
        "Uncaught thread exception",
        exc_info=(args.exc_type, args.exc_value, args.exc_traceback),
    )



from typing import Any, Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class TextState(StringBaseState):
    class Attributes(AttributesBase):
        editable: bool | None = Field(default=None)
        min: int | float | None = Field(default=None)
        max: int | float | None = Field(default=None)
        pattern: Any | None = Field(default=None)
        mode: str | None = Field(default=None)

    domain: Literal["text"]

    attributes: Attributes



from typing import Literal

from pydantic import Field

from hassette.const.sensor import UNIT_OF_MEASUREMENT

from .base import AttributesBase, StringBaseState


class AirQualityState(StringBaseState):
    class Attributes(AttributesBase):
        nitrogen_oxide: float | None = Field(default=None)
        particulate_matter_10: float | None = Field(default=None)
        particulate_matter_2_5: float | None = Field(default=None)
        unit_of_measurement: UNIT_OF_MEASUREMENT | str | None = Field(default=None)
        attribution: str | None = Field(default=None)

    domain: Literal["air_quality"]

    attributes: Attributes



from typing import Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class AssistSatelliteState(StringBaseState):
    class Attributes(AttributesBase):
        restored: bool | None = Field(default=None)

    domain: Literal["assist_satellite"]

    attributes: Attributes



from hassette.core.bus import predicates
from hassette.core.bus.bus import Bus
from hassette.core.bus.listeners import Listener, Subscription

__all__ = [
    "Bus",
    "Listener",
    "Subscription",
    "predicates",
]



import asyncio
import functools
import inspect
from collections.abc import Awaitable, Callable
from logging import getLogger
from typing import ParamSpec, TypeVar, cast, overload

LOGGER = getLogger(__name__)


P = ParamSpec("P")
R = TypeVar("R")


def _is_async_callable(fn: Callable[..., object]) -> bool:
    """True for coroutine functions, including functools.partial(async_fn)."""
    if asyncio.iscoroutinefunction(fn):
        return True
    if isinstance(fn, functools.partial) and asyncio.iscoroutinefunction(fn.func):  # type: ignore[attr-defined]
        return True
    # unwrap decorated callables (keeps metadata nice, too)
    try:
        unwrapped = inspect.unwrap(fn)  # no-op if not wrapped
    except Exception:
        return False
    return asyncio.iscoroutinefunction(unwrapped)


@overload
def make_async_adapter(fn: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]: ...
@overload
def make_async_adapter(fn: Callable[P, R]) -> Callable[P, Awaitable[R]]: ...


def make_async_adapter(fn: Callable[P, R] | Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
    """
    Normalize a callable (sync or async) into an async callable with the same signature.

    - If `fn` is async: await it.
    - If `fn` is sync: run it in Hassette's thread pool executor.
    """

    from hassette.core.core import Hassette

    # Keep original metadata for logging/debugging
    def _wrap(meta_src: Callable[..., object], wrapped: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
        functools.update_wrapper(wrapped, meta_src)  # preserves __name__/__qualname__/__module__/__wrapped__
        return wrapped

    if _is_async_callable(fn):

        async def _async_fn(*args: P.args, **kwargs: P.kwargs) -> R:
            # Cast solely for type-checkers; at runtime it's already awaitable.
            return await cast("Callable[P, Awaitable[R]]", fn)(*args, **kwargs)

        return _wrap(cast("Callable[..., object]", fn), _async_fn)

    # sync function path
    async def _sync_fn(*args: P.args, **kwargs: P.kwargs) -> R:
        inst = Hassette.get_instance()
        loop = inst._loop
        if not loop:
            raise RuntimeError("Event loop is not running")

        # run_in_executor can't take kwargs; use partial to bind both.
        bound = functools.partial(cast("Callable[P, R]", fn), *args, **kwargs)
        fut = loop.run_in_executor(inst._thread_pool, bound)
        try:
            return await fut
        except Exception:
            fut.cancel()
            raise

    return _wrap(cast("Callable[..., object]", fn), _sync_fn)



import typing
from dataclasses import dataclass

if typing.TYPE_CHECKING:
    from hassette.core.events import StateChangeEvent


@dataclass(frozen=True)
class DomainIs:
    domain: str

    def __call__(self, event: "StateChangeEvent") -> bool:
        data = event.payload

        return data.domain == self.domain

    def __repr__(self) -> str:
        return f"DomainIs(domain={self.domain!r})"


@dataclass(frozen=True)
class EntityIs:
    entity_id_wanted: str

    def __call__(self, event: "StateChangeEvent") -> bool:
        data = event.payload.data

        return data.entity_id == self.entity_id_wanted

    def __repr__(self) -> str:
        return f"EntityIs(entity_id_wanted={self.entity_id_wanted!r})"



from logging import getLogger
from typing import Generic, Literal, TypeVar

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from whenever import Date, Instant, PlainDateTime, Time

DomainLiteral = Literal[
    "automation",
    "binary_sensor",
    "button",
    "calendar",
    "climate",
    "conversation",
    "cover",
    "device_tracker",
    "event",
    "fan",
    "humidifier",
    "input_boolean",
    "input_datetime",
    "input_number",
    "input_text",
    "light",
    "media_player",
    "number",
    "person",
    "remote",
    "scene",
    "script",
    "select",
    "sensor",
    "stt",
    "sun",
    "switch",
    "timer",
    "tts",
    "update",
    "weather",
    "zone",
]


StateT = TypeVar("StateT", bound="BaseState")
"""Represents a specific state type, e.g., LightState, Sensor_TemperatureState, etc."""

StateValueT = TypeVar("StateValueT")
"""Represents the type of the state attribute in a State model, e.g. bool for BinarySensorState."""


LOGGER = getLogger(__name__)


class Context(BaseModel):
    model_config = ConfigDict(frozen=True)

    id: str | None = Field(default=None)
    """The context ID of the event."""

    parent_id: str | None = Field(default=None)
    """The parent context ID of the event, if any."""

    user_id: str | None = Field(default=None)
    """The user ID for who triggered the event."""


class AttributesBase(BaseModel):
    """Represents the attributes of a HomeAssistant state."""

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True, coerce_numbers_to_str=True, frozen=True)

    icon: str | None = Field(default=None, repr=False)
    """The icon of the entity."""

    friendly_name: str | None = Field(default=None)
    """A friendly name for the entity."""

    device_class: str | None = Field(default=None)
    """The device class of the entity."""

    entity_id: list[str] | None = Field(default=None)
    """List of entity IDs if this is a group entity."""

    supported_features: int | float | None = Field(default=None)
    """Bitfield of supported features."""


class BaseState(BaseModel, Generic[StateValueT]):
    """Represents a Home Assistant state object."""

    # Note: HA docs mention object_id and name, but I personally haven't seen these in practice.
    # Leaving them off unless we find a use case or get a feature request for them.
    # https://www.home-assistant.io/docs/configuration/state_object/#about-the-state-object

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True, coerce_numbers_to_str=True, frozen=True)

    domain: DomainLiteral | str = Field(...)
    """The domain of the entity, e.g. 'light', 'sensor', etc."""

    entity_id: str = Field(...)
    """The full entity ID, e.g. 'light.living_room'."""

    last_changed: Instant | None = Field(None, repr=False)
    """Time the state changed in the state machine, not updated when only attributes change."""

    last_reported: Instant | None = Field(None, repr=False)
    """Time the state was written to the state machine, updated regardless of any changes to the state or
    state attributes.
    """

    last_updated: Instant | None = Field(None, repr=False)
    """Time the state or state attributes changed in the state machine, not updated if neither state nor state
    attributes changed.
    """

    context: Context = Field(repr=False)
    """The context of the state change."""

    is_unknown: bool = Field(default=False)
    """Whether the state is 'unknown'."""

    is_unavailable: bool = Field(default=False)
    """Whether the state is 'unavailable'."""

    value: StateValueT = Field(..., validation_alias="state")
    """The state value, e.g. 'on', 'off', 23.5, etc."""

    attributes: AttributesBase = Field(...)
    """The attributes of the state."""

    @property
    def is_group(self) -> bool:
        """Whether this entity is a group entity (i.e. has multiple entity_ids)."""
        if not self.attributes:
            return False

        if not hasattr(self.attributes, "entity_id"):
            return False

        if not isinstance(self.attributes.entity_id, list):  # type: ignore
            return False

        return len(self.attributes.entity_id) > 1  # type: ignore

    @model_validator(mode="before")
    @classmethod
    def _validate_domain_and_state(cls, values):
        if not isinstance(values, dict):
            LOGGER.warning("Expected values to be a dict, got %s", type(values).__name__, stacklevel=2)
            return values

        entity_id = values.get("entity_id")
        if entity_id:
            domain = entity_id.split(".")[0]
            values["domain"] = domain

        state = values.get("state")
        if state == "unknown":
            values["is_unknown"] = True
            values["state"] = None
        elif state == "unavailable":
            values["is_unavailable"] = True
            values["state"] = None

        return values


class StringBaseState(BaseState[str | None]):
    """Base class for string states."""


class DateTimeBaseState(BaseState[Instant | PlainDateTime | Date | None]):
    """Base class for datetime states.

    Valid state values are Instant, PlainDateTime, Date, or None.
    """


class TimestampBaseState(BaseState[Instant | PlainDateTime | None]):
    """Base class for timestamp states.

    Valid state values are Instant, PlainDateTime, or None.
    """


class InstantBaseState(BaseState[Instant | None]):
    """Base class for Instant states.

    Valid state values are Instant or None.
    """


class TimeBaseState(BaseState[Time | None]):
    """Base class for Time states.

    Valid state values are Time or None.
    """


class BoolBaseState(BaseState[bool | None]):
    """Base class for boolean states.

    Valids state values are True, False, or None.

    Will convert string values "on" and "off" to boolean True and False.
    """

    @field_validator("value", mode="before")
    @classmethod
    def validate_state(cls, value):
        if value is None:
            return None
        if isinstance(value, str):
            if value.lower() == "on":
                return True
            if value.lower() == "off":
                return False
            raise ValueError(f"Invalid state value: {value}")
        if isinstance(value, bool):
            return value
        raise ValueError(f"State must be a boolean or 'on'/'off' string, got {value}")


class IntBaseState(BaseState[int | None]):
    """Base class for integer states."""

    @field_validator("value", mode="before")
    @classmethod
    def validate_state(cls, value: int | None) -> int | None:
        """Ensure the state value is an integer or None."""
        if value is None:
            return None
        return int(value)


class NumericBaseState(BaseState[float | int | None]):
    """Base class for numeric states.

    Will convert string values to float or int.
    Valid state values are int, float, or None.
    """

    @field_validator("value", mode="before")
    @classmethod
    def validate_state(cls, value: int | float | None) -> int | float | None:
        """Ensure the state value is a number or None."""
        if value is None:
            return None
        if isinstance(value, int | float):
            return value
        return float(value)



from dataclasses import dataclass
from typing import Generic, TypeVar

P = TypeVar("P", covariant=True)


@dataclass(slots=True, frozen=True)
class Event(Generic[P]):
    """Base class for all events, only contains type and payload.

    Payload will be a HassPayload or a HassettePayload depending on the event source."""

    topic: str
    """The topic of the event, used with the Bus to subscribe to specific event types."""

    payload: P
    """The payload of the event, containing the actual event data from HA or Hassette."""



from typing import Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class RemoteState(StringBaseState):
    class Attributes(AttributesBase):
        activity_list: list | None = Field(default=None)
        current_activity: str | None = Field(default=None)

    domain: Literal["remote"]

    attributes: Attributes



import asyncio
import time
import typing
from typing import Any, cast

from hassette.async_utils import make_async_adapter
from hassette.core.events import Event

if typing.TYPE_CHECKING:
    from hassette.core.core import Hassette
    from hassette.core.types import AsyncHandler, E_contra, Handler


def make_async_handler(fn: "Handler[E_contra]") -> "AsyncHandler[E_contra]":
    """Wrap a function to ensure it is always called as an async handler.

    If the function is already an async function, it will be called directly.
    If it is a regular function, it will be run in an executor to avoid blocking the event loop.

    Args:
        fn (Callable[..., Any]): The function to adapt.

    Returns:
        AsyncHandler: An async handler that wraps the original function.
    """
    return cast("AsyncHandler[E_contra]", make_async_adapter(fn))


def add_debounce(
    handler: "AsyncHandler[Event[Any]]", seconds: float, hassette: "Hassette"
) -> "AsyncHandler[Event[Any]]":
    """Add a debounce to an async handler.

    This will ensure that the handler is only called after a specified period of inactivity.
    If a new event comes in before the debounce period has passed, the previous call is cancelled.

    Args:
        handler (AsyncHandler): The async handler to debounce.
        seconds (float): The debounce period in seconds.

    Returns:
        AsyncHandler: A new async handler that applies the debounce logic.
    """
    pending: asyncio.Task | None = None
    last_ev: Event[Any] | None = None

    async def _debounced(event: Event[Any]) -> None:
        nonlocal pending, last_ev
        last_ev = event
        if pending and not pending.done():
            pending.cancel()

        async def _later():
            try:
                await asyncio.sleep(seconds)
                if last_ev is not None:
                    await handler(last_ev)
            except asyncio.CancelledError:
                pass

        pending = hassette.create_task(_later())

    return _debounced


def add_throttle(handler: "AsyncHandler[Event[Any]]", seconds: float) -> "AsyncHandler[Event[Any]]":
    """Add a throttle to an async handler.

    This will ensure that the handler is only called at most once every specified period of time.
    If a new event comes in before the throttle period has passed, it will be ignored.

    Args:
        handler (AsyncHandler): The async handler to throttle.
        seconds (float): The throttle period in seconds.

    Returns:
        AsyncHandler: A new async handler that applies the throttle logic.
    """

    last_time = 0.0
    lock = asyncio.Lock()

    async def _throttled(event: Event[Any]) -> None:
        nonlocal last_time
        async with lock:
            now = time.monotonic()
            if now - last_time >= seconds:
                last_time = now
                await handler(event)

    return _throttled



import asyncio
import contextlib
import heapq
import typing
from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import Any, Generic, TypeVar, cast

from whenever import SystemDateTime, TimeDelta

from hassette.async_utils import make_async_adapter
from hassette.core.classes import Resource, Service
from hassette.core.types import TriggerProtocol

from .triggers import CronTrigger, IntervalTrigger, now

if typing.TYPE_CHECKING:
    from hassette.core.core import Hassette


T = TypeVar("T")


class _Scheduler(Service):
    def __init__(self, hassette: "Hassette"):
        super().__init__(hassette)

        self.min_delay = 0.1
        self.max_delay = 30.0

        self._queue: HeapQueue[ScheduledJob] = HeapQueue()
        self._counter = 0
        self._wakeup_event = asyncio.Event()
        self._exit_event = asyncio.Event()
        self._tasks: set[asyncio.Task] = set()

    async def run_forever(self):
        """Run the scheduler forever, processing jobs as they become due."""
        try:
            await self.handle_start()
            self._exit_event = asyncio.Event()

            while True:
                if self._exit_event.is_set():
                    self.logger.debug("Scheduler exiting")
                    return

                while not self._queue.is_empty() and (peek := self._queue.peek()) and peek.next_run <= now():
                    job = self._queue.pop()
                    self._tasks.add(self.hassette.create_task(self._dispatch_and_log(job)))

                await self._sleep()
        except asyncio.CancelledError:
            self.logger.debug("Scheduler cancelled, stopping")
            await self.handle_stop()
            self._exit_event.set()
        except Exception as e:
            await self.handle_crash(e)
            self._exit_event.set()
            raise
        finally:
            await self._cleanup()

    async def _cleanup(self) -> None:
        """Cleanup resources after the WebSocket connection is closed."""

        # Cancel background tasks
        if self._tasks:
            for task in list(self._tasks):
                if not task.done():
                    task.cancel()
            await asyncio.gather(*self._tasks, return_exceptions=True)
            self.logger.debug("Cancelled %d pending tasks", len(self._tasks))

    def _kick(self):
        """Wake up the scheduler to check for jobs."""
        self._wakeup_event.set()

    async def _sleep(self):
        """Sleep until the next job is due or a kick is received.

        This method will wait for the next job to be due or until a kick is received.
        If a kick is received, it will wake up immediately.
        """
        try:
            await asyncio.wait_for(self._wakeup_event.wait(), timeout=self._get_sleep_time().in_seconds())
            self.logger.debug("Scheduler woke up due to kick")
        except asyncio.CancelledError:
            self.logger.debug("Scheduler sleep cancelled")
            raise
        except TimeoutError:
            self.logger.debug("Scheduler woke up due to timeout")
        finally:
            self._wakeup_event.clear()

    def _get_sleep_time(self) -> TimeDelta:
        """Get the time to sleep until the next job is due.
        If there are no jobs, return a default sleep time.
        """
        if not self._queue.is_empty():
            next_run_time = self._queue.peek_or_raise().next_run
            self.logger.debug("Next job scheduled at %s", next_run_time)
            delay = max((next_run_time - now()).in_seconds(), self.min_delay)
        else:
            delay = 15.0  # or some generous default

        # ensure delay isn't over N seconds
        delay = min(delay, self.max_delay)

        self.logger.debug("Scheduler sleeping for %s seconds", delay)

        return TimeDelta(seconds=delay)

    def _push_job(self, job: "ScheduledJob"):
        """Push a job to the queue."""
        self._queue.push(job)
        self._kick()
        self.logger.debug("Scheduled job: %s", job)

    async def _dispatch_and_log(self, job: "ScheduledJob"):
        """Dispatch a job and log its execution.

        Args:
            job (ScheduledJob): The job to dispatch.
        """
        if job.cancelled:
            self.logger.debug("Job %s is cancelled, skipping dispatch", job)
            return

        self.logger.debug("Dispatching job: %s", job)
        with contextlib.suppress(Exception):
            await self._run_job(job)

        try:
            await self._reschedule_job(job)
        except Exception:
            self.logger.exception("Error rescheduling job %s", job)

    async def _reschedule_job(self, job: "ScheduledJob"):
        """Reschedule a job if it is repeating.

        Args:
            job (ScheduledJob): The job to reschedule.
        """

        if job.cancelled:
            self.logger.debug("Job %s is cancelled, not rescheduling", job)
            return

        if job.repeat and job.trigger:
            curr_next_run = job.next_run
            job.next_run = job.trigger.next_run_time()
            next_run_time_delta = job.next_run - curr_next_run
            assert next_run_time_delta.in_seconds() > 0, "Next run time must be in the future"

            self.logger.debug(
                "Rescheduling repeating job %s from %s to %s (%s)",
                job,
                curr_next_run,
                job.next_run,
                next_run_time_delta.in_seconds(),
            )
            self._push_job(job)

    async def _run_job(self, job: "ScheduledJob"):
        """Run a scheduled job.

        Args:
            job (ScheduledJob): The job to run.
        """

        if job.cancelled:
            self.logger.debug("Job %s is cancelled, skipping", job)
            return

        func = job.job

        run_at_delta = job.next_run - now()
        if run_at_delta.in_seconds() < -1:
            self.logger.warning(
                "Job %s is behind schedule by %s seconds, running now.",
                job,
                abs(run_at_delta.in_seconds()),
            )

        try:
            self.logger.debug("Running job %s at %s", job, now())
            async_func = make_async_adapter(func)
            await async_func()
        except Exception as e:
            self.logger.error("Error running job%s: %s - %s", job, type(e), e)


class Scheduler(Resource):
    def __init__(self, hassette: "Hassette", _scheduler: _Scheduler) -> None:
        super().__init__(hassette)
        self._scheduler = _scheduler

    @property
    def scheduler(self) -> _Scheduler:
        """Get the internal scheduler instance."""
        return self._scheduler

    def add_job(self, job: "ScheduledJob") -> "ScheduledJob":
        """Add a job to the scheduler.

        Args:
            job (ScheduledJob): The job to add.

        Returns:
            ScheduledJob: The added job.
        """

        if not isinstance(job, ScheduledJob):
            raise TypeError(f"Expected ScheduledJob, got {type(job).__name__}")

        self._scheduler._push_job(job)

        return job

    def schedule(
        self,
        func: Callable[[], Awaitable[None]] | Callable[[], None],
        run_at: SystemDateTime,
        trigger: TriggerProtocol | None = None,
        repeat: bool = False,
        name: str = "",
    ) -> "ScheduledJob":
        """Schedule a job to run at a specific time or based on a trigger.

        Args:
            func (Callable): The function to run.
            run_at (SystemDateTime): The time to run the job.
            trigger (TriggerProtocol | None): Optional trigger for repeating jobs.
            repeat (bool): Whether the job should repeat.
            name (str): Optional name for the job.

        Returns:
            ScheduledJob: The scheduled job.
        """

        job = ScheduledJob(next_run=run_at, job=func, trigger=trigger, repeat=repeat, name=name)
        return self.add_job(job)

    def run_once(
        self,
        func: Callable[[], Awaitable[None]] | Callable[[], None],
        run_at: SystemDateTime,
        name: str = "",
    ) -> "ScheduledJob":
        """Schedule a job to run at a specific time.

        Args:
            func (Callable): The function to run.
            run_at (SystemDateTime): The time to run the job.
            name (str): Optional name for the job.

        Returns:
            ScheduledJob: The scheduled job.
        """

        return self.schedule(func, run_at, name=name)

    def run_every(
        self,
        func: Callable[[], Awaitable[None]] | Callable[[], None],
        interval: TimeDelta | float,
        name: str = "",
        start: SystemDateTime | None = None,
    ) -> "ScheduledJob":
        """Schedule a job to run at a fixed interval.

        Args:
            func (Callable): The function to run.
            interval (TimeDelta | float): The interval between runs.
            name (str): Optional name for the job.
            start (SystemDateTime | None): Optional start time for the first run. If provided the job will run at this\
                time. Otherwise it will run at the current time plus the interval.

        Returns:
            ScheduledJob: The scheduled job.
        """

        interval_seconds = interval if isinstance(interval, float | int) else interval.in_seconds()

        first_run = start if start else now().add(seconds=interval_seconds)
        trigger = IntervalTrigger.from_arguments(seconds=interval_seconds, start=first_run)

        return self.schedule(func, first_run, trigger=trigger, repeat=True, name=name)

    def run_in(
        self,
        func: Callable[[], Awaitable[None]] | Callable[[], None | Any],
        delay: TimeDelta | float,
        name: str = "",
        start: SystemDateTime | None = None,
    ) -> "ScheduledJob":
        """Schedule a job to run after a delay.

        Args:
            func (Callable): The function to run.
            delay (TimeDelta | float): The delay before running the job.
            name (str): Optional name for the job.
            start (SystemDateTime | None): Optional start time for the first run. If provided the job will run at this\
                time. Otherwise it will run at the current time plus the delay.

        Returns:
            ScheduledJob: The scheduled job.
        """

        delay_seconds = delay if isinstance(delay, float | int) else delay.in_seconds()

        run_at = start if start else now().add(seconds=delay_seconds)
        return self.schedule(func, run_at, name=name)

    def run_cron(
        self,
        func: Callable[[], Awaitable[None]] | Callable[[], None],
        second: int | str = 0,
        minute: int | str = 0,
        hour: int | str = 0,
        day_of_month: int | str = "*",
        month: int | str = "*",
        day_of_week: int | str = "*",
        name: str = "",
        start: SystemDateTime | None = None,
    ) -> "ScheduledJob":
        """Schedule a job using a cron expression.

        Uses a 6-field format (seconds, minutes, hours, day of month, month, day of week).

        Args:
            func (Callable): The function to run.
            second (int | str): Seconds field of the cron expression.
            minute (int | str): Minutes field of the cron expression.
            hour (int | str): Hours field of the cron expression.
            day_of_month (int | str): Day of month field of the cron expression.
            month (int | str): Month field of the cron expression.
            day_of_week (int | str): Day of week field of the cron expression.
            name (str): Optional name for the job.
            start (SystemDateTime | None): Optional start time for the first run. If provided the job will run at this\
                time. Otherwise it will run at the current time plus the cron schedule.

        Returns:
            ScheduledJob: The scheduled job.
        """
        trigger = CronTrigger.from_arguments(
            second=second,
            minute=minute,
            hour=hour,
            day_of_month=day_of_month,
            month=month,
            day_of_week=day_of_week,
            start=start,
        )
        run_at = trigger.next_run_time()
        return self.schedule(func, run_at, trigger=trigger, repeat=True, name=name)


@dataclass(order=True)
class ScheduledJob:
    """A job scheduled to run based on a trigger or at a specific time."""

    next_run: SystemDateTime
    job: Callable[[], Awaitable[None]] | Callable[[], None] = field(compare=False)
    trigger: TriggerProtocol | None = field(compare=False, default=None)
    repeat: bool = field(compare=False, default=False)
    name: str = field(default="", compare=False)
    cancelled: bool = field(default=False, compare=False)

    def __repr__(self) -> str:
        return f"ScheduledJob(name={self.name!r}, next_run={self.next_run})"

    def __post_init__(self):
        self.next_run = self.next_run.round(unit="second")

        if not self.name:
            self.name = self.job.__name__ if hasattr(self.job, "__name__") else str(self.job)

    def cancel(self) -> None:
        """Cancel the scheduled job by setting the cancelled flag to True."""
        self.cancelled = True


@dataclass
class HeapQueue(Generic[T]):
    _queue: list[T] = field(default_factory=list)

    def push(self, job: T):
        """Push a job onto the queue."""
        heapq.heappush(self._queue, job)

    def pop(self) -> T:
        """Pop the next job from the queue."""
        return heapq.heappop(self._queue)

    def peek(self) -> T | None:
        """Peek at the next job without removing it.

        Returns:
            T | None: The next job in the queue, or None if the queue is empty"""
        return self._queue[0] if self._queue else None

    def peek_or_raise(self) -> T:
        """Peek at the next job without removing it, raising an error if the queue is empty.

        Method that the type checker knows always return a value - call `is_empty` first to avoid exceptions.

        Returns:
            T: The next job in the queue.

        Raises:
            IndexError: If the queue is empty.

        """
        if not self._queue:
            raise IndexError("Peek from an empty queue")
        return cast("T", self.peek())

    def is_empty(self) -> bool:
        """Check if the queue is empty."""
        return not self._queue



from fnmatch import fnmatch

from .listeners import Listener

GLOB_CHARS = "*?["


def add_route(
    exact: dict[str, list[Listener]],
    globs: dict[str, list[Listener]],
    key: str,
    listener: Listener,
) -> None:
    """Add a listener to the appropriate route based on whether it contains glob characters.

    Args:
        exact (dict[str, list[Listener]]): Dictionary for exact matches.
        globs (dict[str, list[Listener]]): Dictionary for glob matches.
        key (str): The topic or key to add the listener to.
        L (Listener): The listener to add.

    """
    if any(ch in key for ch in GLOB_CHARS):
        globs.setdefault(key, []).append(listener)
    else:
        exact.setdefault(key, []).append(listener)


def remove_route(
    exact: dict[str, list[Listener]],
    globs: dict[str, list[Listener]],
    key: str,
    predicate,
) -> None:
    """Remove a listener from the appropriate route based on whether it contains glob characters.

    Args:
        exact (dict[str, list[Listener]]): Dictionary for exact matches.
        globs (dict[str, list[Listener]]): Dictionary for glob matches.
        key (str): The topic or key to remove the listener from.
        predicate (callable): A function that returns True for listeners to be removed."""
    bucket = globs if any(ch in key for ch in GLOB_CHARS) else exact
    if key in bucket:
        bucket[key] = [L for L in bucket[key] if not predicate(L)]
        if not bucket[key]:
            bucket.pop(key, None)


def matching_listeners(
    exact: dict[str, list[Listener]],
    globs: dict[str, list[Listener]],
    topic: str,
) -> list[Listener]:
    """Get all listeners that match the given topic.

    Args:
        exact (dict[str, list[Listener]]): Dictionary for exact matches.
        globs (dict[str, list[Listener]]): Dictionary for glob matches.
        topic (str): The topic to match against.

    Returns:
        list[Listener]: A list of listeners that match the topic.
    """
    out: list[Listener] = []
    out.extend(exact.get(topic, ()))

    for k, lst in globs.items():
        if fnmatch(topic, k):
            out.extend(lst)

    # de-dup preserving order
    seen, unique = set(), []
    for listener in out:
        if id(listener) not in seen:
            seen.add(id(listener))
            unique.append(listener)
    return unique
